@@ -11,16 +11,23 @@ from typing import (
     Set,
     Tuple,
     Union,
+    cast,
 )
 
 from dagster import check
 from dagster.config import Field, Shape
 from dagster.config.config_type import ConfigType
+from dagster.config.validate import process_config
 from dagster.core.definitions.config import ConfigMapping
 from dagster.core.definitions.definition_config_schema import IDefinitionConfigSchema
 from dagster.core.definitions.mode import ModeDefinition
 from dagster.core.definitions.resource import ResourceDefinition
-from dagster.core.errors import DagsterInvalidDefinitionError
+from dagster.core.definitions.utils import check_valid_name
+from dagster.core.errors import (
+    DagsterInvalidConfigError,
+    DagsterInvalidDefinitionError,
+    DagsterInvariantViolationError,
+)
 from dagster.core.storage.io_manager import io_manager
 from dagster.core.types.dagster_type import (
     DagsterType,
@@ -28,38 +35,40 @@ from dagster.core.types.dagster_type import (
     construct_dagster_type_dictionary,
 )
 from dagster.utils import merge_dicts
-from dagster.utils.backcompat import experimental
 from toposort import CircularDependencyError, toposort_flatten
 
 from .dependency import (
     DependencyStructure,
     IDependencyDefinition,
-    Solid,
-    SolidHandle,
+    Node,
+    NodeHandle,
     SolidInputHandle,
     SolidInvocation,
 )
 from .hook import HookDefinition
-from .i_solid_definition import NodeDefinition
 from .input import FanInInputPointer, InputDefinition, InputMapping, InputPointer
 from .logger import LoggerDefinition
+from .node import NodeDefinition
 from .output import OutputDefinition, OutputMapping
 from .preset import PresetDefinition
 from .solid_container import create_execution_structure, validate_dependency_dict
+from .version_strategy import VersionStrategy
 
 if TYPE_CHECKING:
     from dagster.core.instance import DagsterInstance
     from .solid import SolidDefinition
     from .partition import PartitionedConfig
     from .executor import ExecutorDefinition
-    from .pipeline import PipelineDefinition
-    from dagster.core.execution.execute import InProcessGraphResult
+    from .job import JobDefinition
+    from dagster.core.execution.execute_in_process_result import ExecuteInProcessResult
 
 
-def _check_node_defs_arg(graph_name: str, node_defs: List[NodeDefinition]):
+def _check_node_defs_arg(graph_name: str, node_defs: Optional[List[NodeDefinition]]):
+    node_defs = node_defs or []
+
     if not isinstance(node_defs, list):
         raise DagsterInvalidDefinitionError(
-            '"solids" arg to "{name}" is not a list. Got {val}.'.format(
+            '"nodes" arg to "{name}" is not a list. Got {val}.'.format(
                 name=graph_name, val=repr(node_defs)
             )
         )
@@ -69,27 +78,27 @@ def _check_node_defs_arg(graph_name: str, node_defs: List[NodeDefinition]):
         elif callable(node_def):
             raise DagsterInvalidDefinitionError(
                 """You have passed a lambda or function {func} into {name} that is
-                not a solid. You have likely forgetten to annotate this function with
-                an @solid or @lambda_solid decorator.'
+                not a node. You have likely forgetten to annotate this function with
+                the @op or @graph decorators.'
                 """.format(
                     name=graph_name, func=node_def.__name__
                 )
             )
         else:
             raise DagsterInvalidDefinitionError(
-                "Invalid item in solid list: {item}".format(item=repr(node_def))
+                "Invalid item in node list: {item}".format(item=repr(node_def))
             )
 
     return node_defs
 
 
 def _create_adjacency_lists(
-    solids: List[Solid],
+    solids: List[Node],
     dep_structure: DependencyStructure,
-) -> Tuple[Dict[str, Set[Solid]], Dict[str, Set[Solid]]]:
+) -> Tuple[Dict[str, Set[Node]], Dict[str, Set[Node]]]:
     visit_dict = {s.name: False for s in solids}
-    forward_edges: Dict[str, Set[Solid]] = {s.name: set() for s in solids}
-    backward_edges: Dict[str, Set[Solid]] = {s.name: set() for s in solids}
+    forward_edges: Dict[str, Set[Node]] = {s.name: set() for s in solids}
+    backward_edges: Dict[str, Set[Node]] = {s.name: set() for s in solids}
 
     def visit(solid_name):
         if visit_dict[solid_name]:
@@ -115,27 +124,28 @@ class GraphDefinition(NodeDefinition):
     def __init__(
         self,
         name: str,
-        description: Optional[str],
-        node_defs: List[NodeDefinition],
-        dependencies: Optional[Dict[Union[str, SolidInvocation], Dict[str, IDependencyDefinition]]],
-        input_mappings: Optional[List[InputMapping]],
-        output_mappings: Optional[List[OutputMapping]],
-        config_mapping: Optional[ConfigMapping],
+        description: Optional[str] = None,
+        node_defs: Optional[List[NodeDefinition]] = None,
+        dependencies: Optional[
+            Dict[Union[str, SolidInvocation], Dict[str, IDependencyDefinition]]
+        ] = None,
+        input_mappings: Optional[List[InputMapping]] = None,
+        output_mappings: Optional[List[OutputMapping]] = None,
+        config: Optional[ConfigMapping] = None,
+        tags: Optional[Dict[str, Any]] = None,
         **kwargs,
     ):
         self._node_defs = _check_node_defs_arg(name, node_defs)
-        # TODO: backcompat for now
-        self._solid_defs = self._node_defs
         self._dagster_type_dict = construct_dagster_type_dictionary(self._node_defs)
         self._dependencies = validate_dependency_dict(dependencies)
-        self._dependency_structure, self._solid_dict = create_execution_structure(
-            node_defs, self._dependencies, graph_definition=self
+        self._dependency_structure, self._node_dict = create_execution_structure(
+            self._node_defs, self._dependencies, graph_definition=self
         )
 
         # List[InputMapping]
         self._input_mappings, input_defs = _validate_in_mappings(
             check.opt_list_param(input_mappings, "input_mappings"),
-            self._solid_dict,
+            self._node_dict,
             self._dependency_structure,
             name,
             class_name=type(self).__name__,
@@ -143,19 +153,20 @@ class GraphDefinition(NodeDefinition):
         # List[OutputMapping]
         self._output_mappings = _validate_out_mappings(
             check.opt_list_param(output_mappings, "output_mappings"),
-            self._solid_dict,
+            self._node_dict,
             self._dependency_structure,
             name,
             class_name=type(self).__name__,
         )
 
-        self._config_mapping = check.opt_inst_param(config_mapping, "config_mapping", ConfigMapping)
+        self._config_mapping = check.opt_inst_param(config, "config", ConfigMapping)
 
         super(GraphDefinition, self).__init__(
             name=name,
             description=description,
             input_defs=input_defs,
             output_defs=[output_mapping.definition for output_mapping in self._output_mappings],
+            tags=tags,
             **kwargs,
         )
 
@@ -177,12 +188,20 @@ class GraphDefinition(NodeDefinition):
         return [self.solid_named(solid_name) for solid_name in order]
 
     @property
-    def solids(self) -> List[Solid]:
-        return list(set(self._solid_dict.values()))
+    def node_type_str(self) -> str:
+        return "graph"
 
     @property
-    def solid_dict(self) -> Dict[str, Solid]:
-        return self._solid_dict
+    def is_graph_job_op_node(self) -> bool:
+        return True
+
+    @property
+    def solids(self) -> List[Node]:
+        return list(set(self._node_dict.values()))
+
+    @property
+    def node_dict(self) -> Dict[str, Node]:
+        return self._node_dict
 
     @property
     def node_defs(self) -> List[NodeDefinition]:
@@ -190,19 +209,19 @@ class GraphDefinition(NodeDefinition):
 
     def has_solid_named(self, name: str) -> bool:
         check.str_param(name, "name")
-        return name in self._solid_dict
+        return name in self._node_dict
 
-    def solid_named(self, name: str) -> Solid:
+    def solid_named(self, name: str) -> Node:
         check.str_param(name, "name")
         check.invariant(
-            name in self._solid_dict,
+            name in self._node_dict,
             "{graph_name} has no solid named {name}.".format(graph_name=self._name, name=name),
         )
 
-        return self._solid_dict[name]
+        return self._node_dict[name]
 
-    def get_solid(self, handle: SolidHandle) -> Solid:
-        check.inst_param(handle, "handle", SolidHandle)
+    def get_solid(self, handle: NodeHandle) -> Node:
+        check.inst_param(handle, "handle", NodeHandle)
         current = handle
         lineage = []
         while current:
@@ -279,17 +298,17 @@ class GraphDefinition(NodeDefinition):
         check.failed(f"Could not find output mapping {output_name}")
 
     def resolve_output_to_origin(
-        self, output_name: str, handle: SolidHandle
-    ) -> Tuple[OutputDefinition, SolidHandle]:
+        self, output_name: str, handle: NodeHandle
+    ) -> Tuple[OutputDefinition, NodeHandle]:
         check.str_param(output_name, "output_name")
-        check.inst_param(handle, "handle", SolidHandle)
+        check.inst_param(handle, "handle", NodeHandle)
 
         mapping = self.get_output_mapping(output_name)
         check.invariant(mapping, "Can only resolve outputs for valid output names")
         mapped_solid = self.solid_named(mapping.maps_from.solid_name)
         return mapped_solid.definition.resolve_output_to_origin(
             mapping.maps_from.output_name,
-            SolidHandle(mapped_solid.name, handle),
+            NodeHandle(mapped_solid.name, handle),
         )
 
     def default_value_for_input(self, input_name: str) -> Any:
@@ -332,13 +351,18 @@ class GraphDefinition(NodeDefinition):
 
     def input_supports_dynamic_output_dep(self, input_name: str) -> bool:
         mapping = self.get_input_mapping(input_name)
-        internal_dynamic_handle = self.dependency_structure.get_upstream_dynamic_handle_for_solid(
-            mapping.maps_to.solid_name
-        )
-        if internal_dynamic_handle:
+        target_node = mapping.maps_to.solid_name
+        # check if input mapped to solid which is downstream of another dynamic output within
+        if self.dependency_structure.is_dynamic_mapped(target_node):
             return False
 
-        return True
+        # check if input mapped to solid which starts new dynamic downstream
+        if self.dependency_structure.has_dynamic_downstreams(target_node):
+            return False
+
+        return self.solid_named(target_node).definition.input_supports_dynamic_output_dep(
+            mapping.maps_to.input_name
+        )
 
     def copy_for_configured(
         self,
@@ -347,22 +371,43 @@ class GraphDefinition(NodeDefinition):
         config_schema: Any,
         config_or_config_fn: Any,
     ):
-        check.not_implemented("@graph does not yet implement configured")
+        if not self.has_config_mapping:
+            raise DagsterInvalidDefinitionError(
+                "Only graphs utilizing config mapping can be pre-configured. The graph "
+                '"{graph_name}" does not have a config mapping, and thus has nothing to be '
+                "configured.".format(graph_name=self.name)
+            )
+        config_mapping = cast(ConfigMapping, self.config_mapping)
+        return GraphDefinition(
+            name=name,
+            description=check.opt_str_param(description, "description", default=self.description),
+            node_defs=self._node_defs,
+            dependencies=self._dependencies,
+            input_mappings=self._input_mappings,
+            output_mappings=self._output_mappings,
+            config=ConfigMapping(
+                config_mapping.config_fn,
+                config_schema=config_schema,
+                receive_processed_config_values=config_mapping.receive_processed_config_values,
+            ),
+        )
 
     def node_names(self):
-        return list(self._solid_dict.keys())
+        return list(self._node_dict.keys())
 
-    @experimental
     def to_job(
         self,
         name: Optional[str] = None,
+        description: Optional[str] = None,
         resource_defs: Optional[Dict[str, ResourceDefinition]] = None,
         config: Union[ConfigMapping, Dict[str, Any], "PartitionedConfig"] = None,
-        tags: Optional[Dict[str, str]] = None,
+        tags: Optional[Dict[str, Any]] = None,
         logger_defs: Optional[Dict[str, LoggerDefinition]] = None,
         executor_def: Optional["ExecutorDefinition"] = None,
         hooks: Optional[AbstractSet[HookDefinition]] = None,
-    ) -> "PipelineDefinition":
+        version_strategy: Optional[VersionStrategy] = None,
+        op_selection: Optional[List[str]] = None,
+    ) -> "JobDefinition":
         """
         Make this graph in to an executable Job by providing remaining components required for execution.
 
@@ -399,18 +444,25 @@ class GraphDefinition(NodeDefinition):
             logger_defs (Optional[Dict[str, LoggerDefinition]]):
                 A dictionary of string logger identifiers to their implementations.
             executor_def (Optional[ExecutorDefinition]):
-                How this Job will be executed. Defaults to :py:class:`multiprocess_executor` .
+                How this Job will be executed. Defaults to :py:class:`multi_or_in_process_executor`,
+                which can be switched between multi-process and in-process modes of execution. The
+                default mode of execution is multi-process.
+            version_strategy (Optional[VersionStrategy]):
+                Defines how each solid (and optionally, resource) in the job can be versioned. If
+                provided, memoizaton will be enabled for this job.
 
         Returns:
-            PipelineDefinition: The "Job" currently implemented as a single-mode pipeline
+            JobDefinition
         """
-        from .pipeline import PipelineDefinition
+        from .job import JobDefinition
         from .partition import PartitionedConfig
-        from .executor import ExecutorDefinition, multiprocess_executor
+        from .executor import ExecutorDefinition, multi_or_in_process_executor
 
-        tags = check.opt_dict_param(tags, "tags", key_type=str, value_type=str)
+        job_name = check_valid_name(name or self.name)
+
+        tags = check.opt_dict_param(tags, "tags", key_type=str)
         executor_def = check.opt_inst_param(
-            executor_def, "executor_def", ExecutorDefinition, default=multiprocess_executor
+            executor_def, "executor_def", ExecutorDefinition, default=multi_or_in_process_executor
         )
 
         if resource_defs and "io_manager" in resource_defs:
@@ -421,6 +473,7 @@ class GraphDefinition(NodeDefinition):
             )
 
         hooks = check.opt_set_param(hooks, "hooks", of_type=HookDefinition)
+        op_selection = check.opt_list_param(op_selection, "op_selection", of_type=str)
         presets = []
         config_mapping = None
         partitioned_config = None
@@ -434,7 +487,10 @@ class GraphDefinition(NodeDefinition):
             # Using config mapping here is a trick to make it so that the preset will be used even
             # when no config is supplied for the job.
             config_mapping = _config_mapping_with_default_value(
-                self._get_config_schema(resource_defs_with_defaults, executor_def), config
+                self._get_config_schema(resource_defs_with_defaults, executor_def, logger_defs),
+                config,
+                job_name,
+                self.name,
             )
         elif config is not None:
             check.failed(
@@ -442,39 +498,50 @@ class GraphDefinition(NodeDefinition):
                 f"is an object of type {type(config)}"
             )
 
-        job_name = name or self.name
-
-        return PipelineDefinition(
+        return JobDefinition(
             name=job_name,
+            description=description,
             graph_def=self,
-            mode_defs=[
-                ModeDefinition(
-                    resource_defs=resource_defs_with_defaults,
-                    logger_defs=logger_defs,
-                    executor_defs=[executor_def],
-                    _config_mapping=config_mapping,
-                    _partitioned_config=partitioned_config,
-                )
-            ],
+            mode_def=ModeDefinition(
+                resource_defs=resource_defs_with_defaults,
+                logger_defs=logger_defs,
+                executor_defs=[executor_def],
+                _config_mapping=config_mapping,
+                _partitioned_config=partitioned_config,
+            ),
             preset_defs=presets,
             tags=tags,
             hook_defs=hooks,
-        )
+            version_strategy=version_strategy,
+        ).get_job_def_for_op_selection(op_selection)
+
+    def coerce_to_job(self):
+        # attempt to coerce a Graph in to a Job, raising a useful error if it doesn't work
+        try:
+            return self.to_job()
+        except DagsterInvalidDefinitionError as err:
+            raise DagsterInvalidDefinitionError(
+                f"Failed attempting to coerce Graph {self.name} in to a Job. "
+                "Use to_job instead, passing the required information."
+            ) from err
 
     def _get_config_schema(
         self,
         resource_defs: Optional[Dict[str, ResourceDefinition]],
         executor_def: "ExecutorDefinition",
+        logger_defs: Optional[Dict[str, LoggerDefinition]],
     ) -> ConfigType:
-        from .pipeline import PipelineDefinition
+        from .job import JobDefinition
 
         return (
-            PipelineDefinition(
+            JobDefinition(
                 name=self.name,
                 graph_def=self,
-                mode_defs=[
-                    ModeDefinition(resource_defs=resource_defs, executor_defs=[executor_def])
-                ],
+                mode_def=ModeDefinition(
+                    resource_defs=resource_defs,
+                    executor_defs=[executor_def],
+                    logger_defs=logger_defs,
+                ),
             )
             .get_run_config_schema("default")
             .run_config_schema_type
@@ -482,40 +549,64 @@ class GraphDefinition(NodeDefinition):
 
     def execute_in_process(
         self,
-        run_config: Optional[Dict[str, Any]] = None,
+        run_config: Any = None,
         instance: Optional["DagsterInstance"] = None,
         resources: Optional[Dict[str, Any]] = None,
+        raise_on_error: bool = True,
     ):
         """
         Execute this graph in-process, collecting results in-memory.
 
         Args:
             run_config (Optional[Dict[str, Any]]):
-                Configuration for the run.
+                Run config to provide to execution. The configuration for the underlying graph
+                should exist under the "ops" key.
             instance (Optional[DagsterInstance]):
                 The instance to execute against, an ephemeral one will be used if none provided.
             resources (Optional[Dict[str, Any]]):
                 The resources needed if any are required. Can provide resource instances directly,
                 or resource definitions.
+            raise_on_error (Optional[bool]): Whether or not to raise exceptions when they occur.
+                Defaults to ``True``.
 
         Returns:
-            InProcessGraphResult
+            :py:class:`~dagster.ExecuteInProcessResult`
         """
-        from dagster.core.execution.execute import execute_in_process
+        from dagster.core.execution.build_resources import wrap_resources_for_execution
+        from dagster.core.execution.execute_in_process import core_execute_in_process
+        from dagster.core.instance import DagsterInstance
+        from .job import JobDefinition
+        from .executor import execute_in_process_executor
 
-        run_config = check.opt_dict_param(run_config, "run_config")
+        if len(self.input_defs) > 0:
+            raise DagsterInvariantViolationError(
+                "Graphs with inputs cannot be used with execute_in_process at this time."
+            )
 
-        return execute_in_process(
+        instance = check.opt_inst_param(instance, "instance", DagsterInstance)
+        resources = check.opt_dict_param(resources, "resources", key_type=str)
+
+        resource_defs = wrap_resources_for_execution(resources)
+        in_proc_mode = ModeDefinition(
+            executor_defs=[execute_in_process_executor], resource_defs=resource_defs
+        )
+        ephemeral_job = JobDefinition(name=self._name, graph_def=self, mode_def=in_proc_mode)
+
+        run_config = run_config if run_config is not None else {}
+
+        return core_execute_in_process(
             node=self,
+            ephemeral_pipeline=ephemeral_job,
             run_config=run_config,
             instance=instance,
-            resources=resources,
+            output_capturing_enabled=True,
+            raise_on_error=raise_on_error,
         )
 
 
 def _validate_in_mappings(
     input_mappings: List[InputMapping],
-    solid_dict: Dict[str, Solid],
+    solid_dict: Dict[str, Node],
     dependency_structure: DependencyStructure,
     name: str,
     class_name: str,
@@ -547,7 +638,7 @@ def _validate_in_mappings(
         if input_def_dict.get(mapping.definition.name):
             if input_def_dict[mapping.definition.name] != mapping.definition:
                 raise DagsterInvalidDefinitionError(
-                    "In {class_name} {name} multiple input mappings with same "
+                    "In {class_name} '{name}' multiple input mappings with same "
                     "definition name but different definitions".format(
                         name=name, class_name=class_name
                     ),
@@ -577,7 +668,7 @@ def _validate_in_mappings(
         if mapping.maps_to_fan_in:
             if not dependency_structure.has_fan_in_deps(solid_input_handle):
                 raise DagsterInvalidDefinitionError(
-                    'In {class_name} "{name}" input mapping target '
+                    "In {class_name} '{name}' input mapping target "
                     '"{mapping.maps_to.solid_name}.{mapping.maps_to.input_name}" (index {mapping.maps_to.fan_in_index} of fan-in) '
                     "is not a MultiDependencyDefinition.".format(
                         name=name, mapping=mapping, class_name=class_name
@@ -588,7 +679,7 @@ def _validate_in_mappings(
                 inner_deps[mapping.maps_to.fan_in_index] is not MappedInputPlaceholder
             ):
                 raise DagsterInvalidDefinitionError(
-                    'In {class_name} "{name}" input mapping target '
+                    "In {class_name} '{name}' input mapping target "
                     '"{mapping.maps_to.solid_name}.{mapping.maps_to.input_name}" index {mapping.maps_to.fan_in_index} in '
                     "the MultiDependencyDefinition is not a MappedInputPlaceholder".format(
                         name=name, mapping=mapping, class_name=class_name
@@ -604,9 +695,9 @@ def _validate_in_mappings(
         else:
             if dependency_structure.has_deps(solid_input_handle):
                 raise DagsterInvalidDefinitionError(
-                    'In {class_name} "{name}" input mapping target '
+                    "In {class_name} '{name}' input mapping target "
                     '"{mapping.maps_to.solid_name}.{mapping.maps_to.input_name}" '
-                    "is already satisfied by solid output".format(
+                    "is already satisfied by output".format(
                         name=name, mapping=mapping, class_name=class_name
                     )
                 )
@@ -617,7 +708,7 @@ def _validate_in_mappings(
             target_type = target_input.dagster_type
             fan_in_msg = ""
 
-        if target_type != mapping.definition.dagster_type:
+        if target_type != mapping.definition.dagster_type and class_name != "GraphDefinition":
             raise DagsterInvalidDefinitionError(
                 "In {class_name} '{name}' input "
                 "'{mapping.definition.name}' of type {mapping.definition.dagster_type.display_name} maps to "
@@ -654,7 +745,7 @@ def _validate_in_mappings(
 
 def _validate_out_mappings(
     output_mappings: List[OutputMapping],
-    solid_dict: Dict[str, Solid],
+    solid_dict: Dict[str, Node],
     dependency_structure: DependencyStructure,
     name: str,
     class_name: str,
@@ -665,23 +756,28 @@ def _validate_out_mappings(
             target_solid = solid_dict.get(mapping.maps_from.solid_name)
             if target_solid is None:
                 raise DagsterInvalidDefinitionError(
-                    "In {class_name} '{name}' output mapping references solid "
+                    "In {class_name} '{name}' output mapping references node "
                     "'{solid_name}' which it does not contain.".format(
                         name=name, solid_name=mapping.maps_from.solid_name, class_name=class_name
                     )
                 )
             if not target_solid.has_output(mapping.maps_from.output_name):
                 raise DagsterInvalidDefinitionError(
-                    "In {class_name} {name} output mapping from solid '{mapping.maps_from.solid_name}' "
+                    "In {class_name} {name} output mapping from {described_node} "
                     "which contains no output named '{mapping.maps_from.output_name}'".format(
-                        name=name, mapping=mapping, class_name=class_name
+                        described_node=target_solid.describe_node(),
+                        name=name,
+                        mapping=mapping,
+                        class_name=class_name,
                     )
                 )
 
             target_output = target_solid.output_def_named(mapping.maps_from.output_name)
 
-            if mapping.definition.dagster_type.kind != DagsterTypeKind.ANY and (
-                target_output.dagster_type != mapping.definition.dagster_type
+            if (
+                mapping.definition.dagster_type.kind != DagsterTypeKind.ANY
+                and (target_output.dagster_type != mapping.definition.dagster_type)
+                and class_name != "GraphDefinition"
             ):
                 raise DagsterInvalidDefinitionError(
                     "In {class_name} '{name}' output "
@@ -709,7 +805,7 @@ def _validate_out_mappings(
             if dynamic_handle and not mapping.definition.is_dynamic:
                 raise DagsterInvalidDefinitionError(
                     f'In {class_name} "{name}" output "{mapping.definition.name}" mapping from '
-                    f'solid "{mapping.maps_from.solid_name}" must be a DynamicOutputDefinition since it is '
+                    f"{target_solid.describe_node()} must be a DynamicOutputDefinition since it is "
                     f'downstream of dynamic output "{dynamic_handle.describe()}".'
                 )
 
@@ -732,6 +828,8 @@ def _validate_out_mappings(
 def _config_mapping_with_default_value(
     inner_schema: ConfigType,
     default_config: Dict[str, Any],
+    job_name: str,
+    graph_name: str,
 ) -> ConfigMapping:
     if not isinstance(inner_schema, Shape):
         check.failed("Only Shape (dictionary) config_schema allowed on Job ConfigMapping")
@@ -740,6 +838,7 @@ def _config_mapping_with_default_value(
         return x
 
     updated_fields = {}
+    field_aliases = inner_schema.field_aliases
     for name, field in inner_schema.fields.items():
         if name in default_config:
             updated_fields[name] = Field(
@@ -747,13 +846,31 @@ def _config_mapping_with_default_value(
                 default_value=default_config[name],
                 description=field.description,
             )
+        elif name in field_aliases and field_aliases[name] in default_config:
+            updated_fields[name] = Field(
+                config=field.config_type,
+                default_value=default_config[field_aliases[name]],
+                description=field.description,
+            )
+        else:
+            updated_fields[name] = field
+
+    config_schema = Shape(
+        fields=updated_fields,
+        description="run config schema with default values from default_config",
+        field_aliases=inner_schema.field_aliases,
+    )
+
+    config_evr = process_config(config_schema, default_config)
+    if not config_evr.success:
+        raise DagsterInvalidConfigError(
+            f"Error in config when building job '{job_name}' from graph '{graph_name}' ",
+            config_evr.errors,
+            default_config,
+        )
 
     return ConfigMapping(
-        config_fn=config_fn,
-        config_schema=Shape(
-            fields=updated_fields,
-            description="run config schema with default values from default_config",
-        ),
+        config_fn=config_fn, config_schema=config_schema, receive_processed_config_values=False
     )
 
 
