@@ -3,22 +3,21 @@ from contextlib import contextmanager
 from typing import Any, Dict, FrozenSet, Iterator, List, Optional, Tuple, Union
 
 from dagster import check
-from dagster.core.definitions import IPipeline, PipelineDefinition
+from dagster.core.definitions import IPipeline, JobDefinition, PipelineDefinition
 from dagster.core.definitions.pipeline import PipelineSubsetDefinition
 from dagster.core.definitions.pipeline_base import InMemoryPipeline
 from dagster.core.errors import DagsterExecutionInterruptedError, DagsterInvariantViolationError
-from dagster.core.events import DagsterEvent
-from dagster.core.execution.context.system import SystemPipelineExecutionContext
+from dagster.core.events import DagsterEvent, EngineEventData
+from dagster.core.execution.context.system import PlanOrchestrationContext
 from dagster.core.execution.plan.execute_plan import inner_plan_execution_iterator
+from dagster.core.execution.plan.outputs import StepOutputHandle
 from dagster.core.execution.plan.plan import ExecutionPlan
 from dagster.core.execution.plan.state import KnownExecutionState
-from dagster.core.execution.resolve_versions import resolve_memoized_execution_plan
 from dagster.core.execution.retries import RetryMode
-from dagster.core.instance import DagsterInstance, is_memoized_run
+from dagster.core.instance import DagsterInstance
 from dagster.core.selector import parse_step_selection
-from dagster.core.storage.mem_io_manager import InMemoryIOManager
 from dagster.core.storage.pipeline_run import PipelineRun, PipelineRunStatus
-from dagster.core.system_config.objects import EnvironmentConfig
+from dagster.core.system_config.objects import ResolvedRunConfig
 from dagster.core.telemetry import log_repo_stats, telemetry_wrapper
 from dagster.core.utils import str_format_set
 from dagster.utils import merge_dicts
@@ -27,8 +26,9 @@ from dagster.utils.interrupts import capture_interrupts
 
 from .context_creation_pipeline import (
     ExecutionContextManager,
-    PipelineExecutionContextManager,
     PlanExecutionContextManager,
+    PlanOrchestrationContextManager,
+    orchestration_context_event_generator,
     scoped_pipeline_context,
 )
 from .results import PipelineExecutionResult
@@ -54,7 +54,10 @@ from .results import PipelineExecutionResult
 
 
 def execute_run_iterator(
-    pipeline: IPipeline, pipeline_run: PipelineRun, instance: DagsterInstance
+    pipeline: IPipeline,
+    pipeline_run: PipelineRun,
+    instance: DagsterInstance,
+    resume_from_failure: bool = False,
 ) -> Iterator[DagsterEvent]:
     check.inst_param(pipeline, "pipeline", IPipeline)
     check.inst_param(pipeline_run, "pipeline_run", PipelineRun)
@@ -70,13 +73,24 @@ def execute_run_iterator(
 
         return gen_execute_on_cancel()
 
-    check.invariant(
-        pipeline_run.status == PipelineRunStatus.NOT_STARTED
-        or pipeline_run.status == PipelineRunStatus.STARTING,
-        desc="Pipeline run {} ({}) in state {}, expected NOT_STARTED or STARTING".format(
-            pipeline_run.pipeline_name, pipeline_run.run_id, pipeline_run.status
-        ),
-    )
+    if not resume_from_failure:
+        # this should be the first run worker to start
+        check.invariant(
+            pipeline_run.status == PipelineRunStatus.NOT_STARTED
+            or pipeline_run.status == PipelineRunStatus.STARTING,
+            desc="Pipeline run {} ({}) in state {}, expected NOT_STARTED or STARTING".format(
+                pipeline_run.pipeline_name, pipeline_run.run_id, pipeline_run.status
+            ),
+        )
+    else:
+        check.invariant(
+            pipeline_run.status == PipelineRunStatus.STARTED
+            or pipeline_run.status == PipelineRunStatus.STARTING,
+            desc="Pipeline run {} ({}) in state {}, expected STARTED or STARTING because it's "
+            "resuming from a run worker failure".format(
+                pipeline_run.pipeline_name, pipeline_run.run_id, pipeline_run.status
+            ),
+        )
 
     if pipeline_run.solids_to_execute:
         pipeline_def = pipeline.get_definition()
@@ -97,29 +111,23 @@ def execute_run_iterator(
                 pipeline_run.solids_to_execute
             )
 
-    known_state = None
-    if pipeline_run.parent_run_id and pipeline_run.step_keys_to_execute:
-        plan_snap = instance.get_execution_plan_snapshot(pipeline_run.execution_plan_snapshot_id)
-        known_state = plan_snap.initial_known_state
-
-    execution_plan = create_execution_plan(
-        pipeline,
-        run_config=pipeline_run.run_config,
-        mode=pipeline_run.mode,
-        step_keys_to_execute=pipeline_run.step_keys_to_execute,
-        known_state=known_state,
-    )
+    execution_plan = _get_execution_plan_from_run(pipeline, pipeline_run, instance)
 
     return iter(
         ExecuteRunWithPlanIterable(
             execution_plan=execution_plan,
             iterator=pipeline_execution_iterator,
-            execution_context_manager=PipelineExecutionContextManager(
+            execution_context_manager=PlanOrchestrationContextManager(
+                context_event_generator=orchestration_context_event_generator,
+                pipeline=pipeline,
                 execution_plan=execution_plan,
                 pipeline_run=pipeline_run,
                 instance=instance,
                 run_config=pipeline_run.run_config,
                 raise_on_error=False,
+                executor_defs=None,
+                output_capture=None,
+                resume_from_failure=resume_from_failure,
             ),
         )
     )
@@ -146,11 +154,16 @@ def execute_run(
         PipelineExecutionResult: The result of the execution.
     """
     if isinstance(pipeline, PipelineDefinition):
+        if isinstance(pipeline, JobDefinition):
+            error = "execute_run requires a reconstructable job but received job definition directly instead."
+        else:
+            error = (
+                "execute_run requires a reconstructable pipeline but received pipeline definition "
+                "directly instead."
+            )
         raise DagsterInvariantViolationError(
-            "execute_run requires an IPipeline but received a PipelineDefinition "
-            "directly instead. To support hand-off to other processes provide a "
-            "ReconstructablePipeline which can be done using reconstructable(). For in "
-            "process only execution you can use InMemoryPipeline."
+            f"{error} To support hand-off to other processes please wrap your definition in "
+            "a call to reconstructable(). Learn more about reconstructable here: https://docs.dagster.io/_apidocs/execution#dagster.reconstructable"
         )
 
     check.inst_param(pipeline, "pipeline", IPipeline)
@@ -191,64 +204,39 @@ def execute_run(
                 pipeline_run.solids_to_execute
             )
 
-    known_state = None
-    if pipeline_run.parent_run_id and pipeline_run.step_keys_to_execute:
-        plan_snap = instance.get_execution_plan_snapshot(pipeline_run.execution_plan_snapshot_id)
-        known_state = plan_snap.initial_known_state
+    execution_plan = _get_execution_plan_from_run(pipeline, pipeline_run, instance)
 
-    execution_plan = create_execution_plan(
-        pipeline,
-        run_config=pipeline_run.run_config,
-        mode=pipeline_run.mode,
-        step_keys_to_execute=pipeline_run.step_keys_to_execute,
-        known_state=known_state,
-    )
-
-    if is_memoized_run(pipeline_run.tags):
-        execution_plan = resolve_memoized_execution_plan(
-            execution_plan, pipeline_run.run_config, instance
-        )
+    output_capture: Optional[Dict[StepOutputHandle, Any]] = {}
 
     _execute_run_iterable = ExecuteRunWithPlanIterable(
         execution_plan=execution_plan,
         iterator=pipeline_execution_iterator,
-        execution_context_manager=PipelineExecutionContextManager(
+        execution_context_manager=PlanOrchestrationContextManager(
+            context_event_generator=orchestration_context_event_generator,
+            pipeline=pipeline,
             execution_plan=execution_plan,
             pipeline_run=pipeline_run,
             instance=instance,
             run_config=pipeline_run.run_config,
             raise_on_error=raise_on_error,
+            executor_defs=None,
+            output_capture=output_capture,
         ),
     )
     event_list = list(_execute_run_iterable)
-    pipeline_context = _execute_run_iterable.pipeline_context
-
-    # workaround for mem_io_manager to work in reconstruct_context, e.g. result.result_for_solid
-    # in-memory values dict will get lost when the resource is re-initiated in reconstruct_context
-    # so instead of re-initiating every single resource, we pass the resource instances to
-    # reconstruct_context directly to avoid re-building from resource def.
-    resource_instances_to_override = {}
-    if pipeline_context:  # None if we have a pipeline failure
-        for (
-            key,
-            resource_instance,
-        ) in pipeline_context.scoped_resources_builder.resource_instance_dict.items():
-            if isinstance(resource_instance, InMemoryIOManager):
-                resource_instances_to_override[key] = resource_instance
 
     return PipelineExecutionResult(
         pipeline.get_definition(),
         pipeline_run.run_id,
         event_list,
-        lambda hardcoded_resources_arg: scoped_pipeline_context(
+        lambda: scoped_pipeline_context(
             execution_plan,
+            pipeline,
             pipeline_run.run_config,
             pipeline_run,
             instance,
-            intermediate_storage=pipeline_context.intermediate_storage,
-            resource_instances_to_override=hardcoded_resources_arg,
         ),
-        resource_instances_to_override=resource_instances_to_override,
+        output_capture=output_capture,
     )
 
 
@@ -272,7 +260,7 @@ def execute_pipeline_iterator(
 
     Parameters:
         pipeline (Union[IPipeline, PipelineDefinition]): The pipeline to execute.
-        run_config (Optional[dict]): The environment configuration that parametrizes this run,
+        run_config (Optional[dict]): The configuration that parametrizes this run,
             as a dict.
         mode (Optional[str]): The name of the pipeline mode to use. You may not set both ``mode``
             and ``preset``.
@@ -353,7 +341,7 @@ def execute_pipeline(
 
     Parameters:
         pipeline (Union[IPipeline, PipelineDefinition]): The pipeline to execute.
-        run_config (Optional[dict]): The environment configuration that parametrizes this run,
+        run_config (Optional[dict]): The configuration that parametrizes this run,
             as a dict.
         mode (Optional[str]): The name of the pipeline mode to use. You may not set both ``mode``
             and ``preset``.
@@ -433,7 +421,12 @@ def _logged_execute_pipeline(
         tags=tags,
     )
 
-    return execute_run(pipeline, pipeline_run, instance, raise_on_error=raise_on_error)
+    return execute_run(
+        pipeline,
+        pipeline_run,
+        instance,
+        raise_on_error=raise_on_error,
+    )
 
 
 def reexecute_pipeline(
@@ -456,7 +449,7 @@ def reexecute_pipeline(
         pipeline (Union[IPipeline, PipelineDefinition]): The pipeline to execute.
         parent_run_id (str): The id of the previous run to reexecute. The run must exist in the
             instance.
-        run_config (Optional[dict]): The environment configuration that parametrizes this run,
+        run_config (Optional[dict]): The configuration that parametrizes this run,
             as a dict.
         solid_selection (Optional[List[str]]): A list of solid selection queries (including single
             solid names) to execute. For example:
@@ -499,18 +492,17 @@ def reexecute_pipeline(
         )
 
         parent_pipeline_run = execute_instance.get_run_by_id(parent_run_id)
-        check.invariant(
-            parent_pipeline_run,
-            "No parent run with id {parent_run_id} found in instance.".format(
-                parent_run_id=parent_run_id
-            ),
-        )
+        if parent_pipeline_run is None:
+            check.failed(
+                "No parent run with id {parent_run_id} found in instance.".format(
+                    parent_run_id=parent_run_id
+                ),
+            )
 
-        step_keys_to_execute: Optional[List[str]] = None
         execution_plan: Optional[ExecutionPlan] = None
         # resolve step selection DSL queries using parent execution information
         if step_selection:
-            step_keys_to_execute, execution_plan = _resolve_reexecute_step_selection(
+            execution_plan = _resolve_reexecute_step_selection(
                 execute_instance,
                 pipeline,
                 mode,
@@ -527,8 +519,6 @@ def reexecute_pipeline(
             tags=tags,
             solid_selection=parent_pipeline_run.solid_selection,
             solids_to_execute=parent_pipeline_run.solids_to_execute,
-            # convert to frozenset https://github.com/dagster-io/dagster/issues/2914
-            step_keys_to_execute=list(step_keys_to_execute) if step_keys_to_execute else None,
             root_run_id=parent_pipeline_run.root_run_id or parent_pipeline_run.run_id,
             parent_run_id=parent_pipeline_run.run_id,
         )
@@ -564,7 +554,7 @@ def reexecute_pipeline_iterator(
         pipeline (Union[IPipeline, PipelineDefinition]): The pipeline to execute.
         parent_run_id (str): The id of the previous run to reexecute. The run must exist in the
             instance.
-        run_config (Optional[dict]): The environment configuration that parametrizes this run,
+        run_config (Optional[dict]): The configuration that parametrizes this run,
             as a dict.
         solid_selection (Optional[List[str]]): A list of solid selection queries (including single
             solid names) to execute. For example:
@@ -603,18 +593,17 @@ def reexecute_pipeline_iterator(
             solid_selection=None,
         )
         parent_pipeline_run = execute_instance.get_run_by_id(parent_run_id)
-        check.invariant(
-            parent_pipeline_run,
-            "No parent run with id {parent_run_id} found in instance.".format(
-                parent_run_id=parent_run_id
-            ),
-        )
+        if parent_pipeline_run is None:
+            check.failed(
+                "No parent run with id {parent_run_id} found in instance.".format(
+                    parent_run_id=parent_run_id
+                ),
+            )
 
-        step_keys_to_execute: Optional[List[str]] = None
         execution_plan: Optional[ExecutionPlan] = None
         # resolve step selection DSL queries using parent execution information
         if step_selection:
-            step_keys_to_execute, execution_plan = _resolve_reexecute_step_selection(
+            execution_plan = _resolve_reexecute_step_selection(
                 execute_instance,
                 pipeline,
                 mode,
@@ -631,8 +620,6 @@ def reexecute_pipeline_iterator(
             tags=tags,
             solid_selection=parent_pipeline_run.solid_selection,
             solids_to_execute=parent_pipeline_run.solids_to_execute,
-            # convert to frozenset https://github.com/dagster-io/dagster/issues/2914
-            step_keys_to_execute=list(step_keys_to_execute) if step_keys_to_execute else None,
             root_run_id=parent_pipeline_run.root_run_id or parent_pipeline_run.run_id,
             parent_run_id=parent_pipeline_run.run_id,
         )
@@ -642,12 +629,14 @@ def reexecute_pipeline_iterator(
 
 def execute_plan_iterator(
     execution_plan: ExecutionPlan,
+    pipeline: IPipeline,
     pipeline_run: PipelineRun,
     instance: DagsterInstance,
     retry_mode: Optional[RetryMode] = None,
     run_config: Optional[dict] = None,
 ) -> Iterator[DagsterEvent]:
     check.inst_param(execution_plan, "execution_plan", ExecutionPlan)
+    check.inst_param(pipeline, "pipeline", IPipeline)
     check.inst_param(pipeline_run, "pipeline_run", PipelineRun)
     check.inst_param(instance, "instance", DagsterInstance)
     retry_mode = check.opt_inst_param(retry_mode, "retry_mode", RetryMode, RetryMode.DISABLED)
@@ -658,12 +647,12 @@ def execute_plan_iterator(
             execution_plan=execution_plan,
             iterator=inner_plan_execution_iterator,
             execution_context_manager=PlanExecutionContextManager(
+                pipeline=pipeline,
                 retry_mode=retry_mode,
                 execution_plan=execution_plan,
                 run_config=run_config,
                 pipeline_run=pipeline_run,
                 instance=instance,
-                raise_on_error=False,
             ),
         )
     )
@@ -671,6 +660,7 @@ def execute_plan_iterator(
 
 def execute_plan(
     execution_plan: ExecutionPlan,
+    pipeline: IPipeline,
     instance: DagsterInstance,
     pipeline_run: PipelineRun,
     run_config: Optional[Dict] = None,
@@ -680,6 +670,7 @@ def execute_plan(
     execute_pipeline() above.
     """
     check.inst_param(execution_plan, "execution_plan", ExecutionPlan)
+    check.inst_param(pipeline, "pipeline", IPipeline)
     check.inst_param(instance, "instance", DagsterInstance)
     check.inst_param(pipeline_run, "pipeline_run", PipelineRun)
     run_config = check.opt_dict_param(run_config, "run_config")
@@ -688,6 +679,7 @@ def execute_plan(
     return list(
         execute_plan_iterator(
             execution_plan=execution_plan,
+            pipeline=pipeline,
             run_config=run_config,
             pipeline_run=pipeline_run,
             instance=instance,
@@ -705,45 +697,74 @@ def _check_pipeline(pipeline: Union[PipelineDefinition, IPipeline]) -> IPipeline
     return pipeline
 
 
+def _get_execution_plan_from_run(
+    pipeline: IPipeline, pipeline_run: PipelineRun, instance: DagsterInstance
+) -> ExecutionPlan:
+    if (
+        # need to rebuild execution plan so it matches the subsetted graph
+        pipeline.solids_to_execute is None
+        and pipeline_run.execution_plan_snapshot_id
+    ):
+        execution_plan_snapshot = instance.get_execution_plan_snapshot(
+            pipeline_run.execution_plan_snapshot_id
+        )
+        if execution_plan_snapshot.can_reconstruct_plan:
+            return ExecutionPlan.rebuild_from_snapshot(
+                pipeline_run.pipeline_name,
+                execution_plan_snapshot,
+            )
+    return create_execution_plan(
+        pipeline,
+        run_config=pipeline_run.run_config,
+        mode=pipeline_run.mode,
+        step_keys_to_execute=pipeline_run.step_keys_to_execute,
+    )
+
+
 def create_execution_plan(
     pipeline: Union[IPipeline, PipelineDefinition],
     run_config: Optional[dict] = None,
     mode: Optional[str] = None,
     step_keys_to_execute: Optional[List[str]] = None,
     known_state: KnownExecutionState = None,
+    instance: Optional[DagsterInstance] = None,
+    tags: Optional[Dict[str, str]] = None,
 ) -> ExecutionPlan:
     pipeline = _check_pipeline(pipeline)
     pipeline_def = pipeline.get_definition()
     check.inst_param(pipeline_def, "pipeline_def", PipelineDefinition)
     run_config = check.opt_dict_param(run_config, "run_config", key_type=str)
     mode = check.opt_str_param(mode, "mode", default=pipeline_def.get_default_mode_name())
-    check.opt_list_param(step_keys_to_execute, "step_keys_to_execute", of_type=str)
+    check.opt_nullable_list_param(step_keys_to_execute, "step_keys_to_execute", of_type=str)
+    check.opt_inst_param(instance, "instance", DagsterInstance)
+    tags = check.opt_dict_param(tags, "tags", key_type=str, value_type=str)
 
-    environment_config = EnvironmentConfig.build(pipeline_def, run_config, mode=mode)
+    resolved_run_config = ResolvedRunConfig.build(pipeline_def, run_config, mode=mode)
 
     return ExecutionPlan.build(
         pipeline,
-        environment_config,
-        mode=mode,
+        resolved_run_config,
         step_keys_to_execute=step_keys_to_execute,
         known_state=known_state,
+        instance=instance,
+        tags=tags,
     )
 
 
 def pipeline_execution_iterator(
-    pipeline_context: SystemPipelineExecutionContext, execution_plan: ExecutionPlan
+    pipeline_context: PlanOrchestrationContext, execution_plan: ExecutionPlan
 ) -> Iterator[DagsterEvent]:
     """A complete execution of a pipeline. Yields pipeline start, success,
     and failure events.
 
     Args:
-        pipeline_context (SystemPipelineExecutionContext):
+        pipeline_context (PlanOrchestrationContext):
         execution_plan (ExecutionPlan):
     """
-    check.inst_param(pipeline_context, "pipeline_context", SystemPipelineExecutionContext)
-    check.inst_param(execution_plan, "execution_plan", ExecutionPlan)
 
-    yield DagsterEvent.pipeline_start(pipeline_context)
+    # TODO: restart event?
+    if not pipeline_context.resume_from_failure:
+        yield DagsterEvent.pipeline_start(pipeline_context)
 
     pipeline_exception_info = None
     pipeline_canceled_info = None
@@ -760,25 +781,36 @@ def pipeline_execution_iterator(
         # (see https://amir.rachum.com/blog/2017/03/03/generator-cleanup/).
         generator_closed = True
         pipeline_exception_info = serializable_error_info_from_exc_info(sys.exc_info())
-        raise
+        if pipeline_context.raise_on_error:
+            raise
     except (KeyboardInterrupt, DagsterExecutionInterruptedError):
         pipeline_canceled_info = serializable_error_info_from_exc_info(sys.exc_info())
-        raise
+        if pipeline_context.raise_on_error:
+            raise
     except Exception:  # pylint: disable=broad-except
         pipeline_exception_info = serializable_error_info_from_exc_info(sys.exc_info())
-        raise  # finally block will run before this is re-raised
+        if pipeline_context.raise_on_error:
+            raise  # finally block will run before this is re-raised
     finally:
         if pipeline_canceled_info:
             reloaded_run = pipeline_context.instance.get_run_by_id(pipeline_context.run_id)
             if reloaded_run and reloaded_run.status == PipelineRunStatus.CANCELING:
                 event = DagsterEvent.pipeline_canceled(pipeline_context, pipeline_canceled_info)
             else:
-                event = DagsterEvent.pipeline_failure(
-                    pipeline_context,
-                    "Execution was interrupted unexpectedly. "
-                    "No user initiated termination request was found, treating as failure.",
-                    pipeline_canceled_info,
-                )
+                if pipeline_context.instance.run_will_resume(pipeline_context.run_id):
+                    event = DagsterEvent.engine_event(
+                        pipeline_context,
+                        "Execution was interrupted unexpectedly. "
+                        "No user initiated termination request was found, not treating as failure because run will be resumed.",
+                        EngineEventData(),
+                    )
+                else:
+                    event = DagsterEvent.pipeline_failure(
+                        pipeline_context,
+                        "Execution was interrupted unexpectedly. "
+                        "No user initiated termination request was found, treating as failure.",
+                        pipeline_canceled_info,
+                    )
         elif pipeline_exception_info:
             event = DagsterEvent.pipeline_failure(
                 pipeline_context,
@@ -954,7 +986,10 @@ def _resolve_reexecute_step_selection(
     run_config: Optional[dict],
     parent_pipeline_run: PipelineRun,
     step_selection: List[str],
-) -> Tuple[List[str], ExecutionPlan]:
+) -> ExecutionPlan:
+    if parent_pipeline_run.solid_selection:
+        pipeline = pipeline.subset_for_execution(parent_pipeline_run.solid_selection)
+
     parent_logs = instance.all_logs(parent_pipeline_run.run_id)
     parent_plan = create_execution_plan(
         pipeline,
@@ -967,6 +1002,8 @@ def _resolve_reexecute_step_selection(
         pipeline,
         run_config,
         mode,
+        step_keys_to_execute=list(step_keys_to_execute),
         known_state=KnownExecutionState.for_reexecution(parent_logs, step_keys_to_execute),
+        tags=parent_pipeline_run.tags,
     )
-    return step_keys_to_execute, execution_plan
+    return execution_plan
