@@ -8,13 +8,13 @@ from dagster.core.errors import (
     DagsterUnknownStepStateError,
 )
 from dagster.core.events import DagsterEvent
-from dagster.core.execution.context.system import SystemPipelineExecutionContext
+from dagster.core.execution.context.system import PlanOrchestrationContext
 from dagster.core.execution.plan.state import KnownExecutionState
 from dagster.core.execution.retries import RetryMode, RetryState
 from dagster.core.storage.tags import PRIORITY_TAG
 from dagster.utils.interrupts import pop_captured_interrupt
 
-from .outputs import StepOutputHandle
+from .outputs import StepOutputData, StepOutputHandle
 from .plan import ExecutionPlan
 from .step import ExecutionStep
 
@@ -56,16 +56,19 @@ class ActiveExecution:
         self._pending: Dict[str, Set[str]] = self._plan.get_executable_step_deps()
 
         # track mapping keys from DynamicOutputs, step_key, output_name -> list of keys
+        # to _gathering while in flight
+        self._gathering_dynamic_outputs: Dict[str, Dict[str, List[str]]] = {}
+        # then on success move to _successful
         self._successful_dynamic_outputs: Dict[str, Dict[str, List[str]]] = (
             dict(self._plan.known_state.dynamic_mappings) if self._plan.known_state else {}
         )
+        self._new_dynamic_mappings: bool = False
 
         # steps move in to these buckets as a result of _update calls
         self._executable: List[str] = []
         self._pending_skip: List[str] = []
         self._pending_retry: List[str] = []
         self._pending_abandon: List[str] = []
-        self._pending_resolve: List[str] = []
         self._waiting_to_retry: Dict[str, float] = {}
 
         # then are considered _in_flight when vended via get_steps_to_*
@@ -100,35 +103,40 @@ class ActiveExecution:
             pending_action = (
                 self._executable + self._pending_abandon + self._pending_retry + self._pending_skip
             )
-            raise DagsterInvariantViolationError(
-                "Execution of pipeline finished without completing the execution plan,."
-                "{pending_str}{in_flight_str}{action_str}{retry_str}".format(
-                    in_flight_str="\nSteps still in flight: {}".format(self._in_flight)
-                    if self._in_flight
-                    else "",
-                    pending_str="\nSteps pending processing: {}".format(self._pending.keys())
-                    if self._pending
-                    else "",
-                    action_str="\nSteps pending action: {}".format(pending_action)
-                    if pending_action
-                    else "",
-                    retry_str="\nSteps waiting to retry: {}".format(self._waiting_to_retry.keys())
-                    if self._waiting_to_retry
-                    else "",
-                )
+            state_str = "{pending_str}{in_flight_str}{action_str}{retry_str}".format(
+                in_flight_str="\nSteps still in flight: {}".format(self._in_flight)
+                if self._in_flight
+                else "",
+                pending_str="\nSteps pending processing: {}".format(self._pending.keys())
+                if self._pending
+                else "",
+                action_str="\nSteps pending action: {}".format(pending_action)
+                if pending_action
+                else "",
+                retry_str="\nSteps waiting to retry: {}".format(self._waiting_to_retry.keys())
+                if self._waiting_to_retry
+                else "",
             )
+            if self._interrupted:
+                raise DagsterExecutionInterruptedError(
+                    f"Execution was interrupted before completing the execution plan. {state_str}"
+                )
+            else:
+                raise DagsterInvariantViolationError(
+                    f"Execution finished without completing the execution plan. {state_str}"
+                )
 
         # See verify_complete - steps for which we did not observe a failure/success event are in an unknown
         # state so we raise to ensure pipeline failure.
         if len(self._unknown_state) > 0:
             if self._interrupted:
                 raise DagsterExecutionInterruptedError(
-                    "Execution of pipeline exited with steps {step_list} in an unknown state after "
+                    "Execution exited with steps {step_list} in an unknown state after "
                     "being interrupted.".format(step_list=self._unknown_state)
                 )
             else:
                 raise DagsterUnknownStepStateError(
-                    "Execution of pipeline exited with steps {step_list} in an unknown state to this process.\n"
+                    "Execution exited with steps {step_list} in an unknown state to this process.\n"
                     "This was likely caused by losing communication with the process performing step execution.".format(
                         step_list=self._unknown_state
                     )
@@ -145,15 +153,12 @@ class ActiveExecution:
         successful_or_skipped_steps = self._success | self._skipped
         failed_or_abandoned_steps = self._failed | self._abandoned
 
-        # make a copy since we are mutating during iteration
-        pending_resolve_snapshot = list(self._pending_resolve)
-        for idx, key in enumerate(pending_resolve_snapshot):
-            new_step_deps = self._plan.resolve(key, self._successful_dynamic_outputs[key])
+        if self._new_dynamic_mappings:
+            new_step_deps = self._plan.resolve(self._successful_dynamic_outputs)
             for step_key, deps in new_step_deps.items():
-                # does this work right with step_keys_to_execute?
                 self._pending[step_key] = deps
 
-            del self._pending_resolve[idx]
+            self._new_dynamic_mappings = False
 
         for step_key, requirements in self._pending.items():
             # If any upstream deps failed - this is not executable
@@ -264,9 +269,11 @@ class ActiveExecution:
         steps = []
         steps_to_skip = list(self._pending_skip)
         for key in steps_to_skip:
-            steps.append(self.get_step_by_key(key))
+            step = self.get_step_by_key(key)
+            steps.append(step)
             self._in_flight.add(key)
             self._pending_skip.remove(key)
+            self._prep_for_dynamic_outputs(step)
 
         return sorted(steps, key=self._sort_key_fn)
 
@@ -338,12 +345,12 @@ class ActiveExecution:
     def mark_success(self, step_key: str) -> None:
         self._success.add(step_key)
         self._mark_complete(step_key)
-        if step_key in self._successful_dynamic_outputs:
-            self._pending_resolve.append(step_key)
+        self._resolve_any_dynamic_outputs(step_key)
 
     def mark_skipped(self, step_key: str) -> None:
         self._skipped.add(step_key)
         self._mark_complete(step_key)
+        self._resolve_any_dynamic_outputs(step_key)
 
     def mark_abandoned(self, step_key: str) -> None:
         self._abandoned.add(step_key)
@@ -389,29 +396,29 @@ class ActiveExecution:
     def handle_event(self, dagster_event: DagsterEvent) -> None:
         check.inst_param(dagster_event, "dagster_event", DagsterEvent)
 
+        step_key = cast(str, dagster_event.step_key)
         if dagster_event.is_step_failure:
-            self.mark_failed(dagster_event.step_key)
+            self.mark_failed(step_key)
         elif dagster_event.is_step_success:
-            self.mark_success(dagster_event.step_key)
+            self.mark_success(step_key)
         elif dagster_event.is_step_skipped:
-            self.mark_skipped(dagster_event.step_key)
+            self.mark_skipped(step_key)
         elif dagster_event.is_step_up_for_retry:
             self.mark_up_for_retry(
-                dagster_event.step_key,
+                step_key,
                 time.time() + dagster_event.step_retry_data.seconds_to_wait
                 if dagster_event.step_retry_data.seconds_to_wait
                 else None,
             )
         elif dagster_event.is_successful_output:
-            self.mark_step_produced_output(dagster_event.event_specific_data.step_output_handle)
+            event_specific_data = cast(StepOutputData, dagster_event.event_specific_data)
+            self.mark_step_produced_output(event_specific_data.step_output_handle)
             if dagster_event.step_output_data.step_output_handle.mapping_key:
-                self._successful_dynamic_outputs[dagster_event.step_key][
+                self._gathering_dynamic_outputs[step_key][
                     dagster_event.step_output_data.step_output_handle.output_name
                 ].append(dagster_event.step_output_data.step_output_handle.mapping_key)
 
-    def verify_complete(
-        self, pipeline_context: SystemPipelineExecutionContext, step_key: str
-    ) -> None:
+    def verify_complete(self, pipeline_context: PlanOrchestrationContext, step_key: str) -> None:
         """Ensure that a step has reached a terminal state, if it has not mark it as an unexpected failure"""
         if step_key in self._in_flight:
             pipeline_context.log.error(
@@ -458,4 +465,26 @@ class ActiveExecution:
         dyn_outputs = [step_out for step_out in step.step_outputs if step_out.is_dynamic]
 
         if dyn_outputs:
-            self._successful_dynamic_outputs[step.key] = {out.name: [] for out in dyn_outputs}
+            self._gathering_dynamic_outputs[step.key] = {out.name: [] for out in dyn_outputs}
+
+    def _resolve_any_dynamic_outputs(self, step_key: str):
+        if step_key in self._gathering_dynamic_outputs:
+            self._successful_dynamic_outputs[step_key] = self._gathering_dynamic_outputs[step_key]
+            self._new_dynamic_mappings = True
+
+    def rebuild_from_events(self, dagster_events: List[DagsterEvent]) -> List[ExecutionStep]:
+        """
+        Replay events to rebuild the execution state and continue after a failure.
+
+        Returns a list of steps that are possibly in flight. Current status of the event log implies
+        that the previous run worker might have crashed before launching these steps, or it may have
+        launched them but they have yet to report a STEP_START event.
+        """
+
+        self.get_steps_to_execute()
+
+        for event in dagster_events:
+            self.handle_event(event)
+            self.get_steps_to_execute()
+
+        return [self.get_step_by_key(step_key) for step_key in self._in_flight]
