@@ -17,9 +17,10 @@ from dagster.core.definitions import (
     GraphDefinition,
     IPipeline,
     InputDefinition,
-    Solid,
+    JobDefinition,
+    Node,
+    NodeHandle,
     SolidDefinition,
-    SolidHandle,
     SolidOutputHandle,
 )
 from dagster.core.definitions.composition import MappedInputPlaceholder
@@ -44,6 +45,8 @@ from dagster.core.system_config.objects import ResolvedRunConfig
 from dagster.core.types.dagster_type import DagsterTypeKind
 from dagster.core.utils import toposort
 
+from ..context.output import get_output_context
+from ..resolve_versions import resolve_step_output_versions
 from .compute import create_step_outputs
 from .inputs import (
     FromConfig,
@@ -74,6 +77,9 @@ if TYPE_CHECKING:
 
 StepHandleTypes = (StepHandle, UnresolvedStepHandle, ResolvedFromDynamicStepHandle)
 StepHandleUnion = Union[StepHandle, UnresolvedStepHandle, ResolvedFromDynamicStepHandle]
+ExecutionStepUnion = Union[
+    ExecutionStep, UnresolvedCollectExecutionStep, UnresolvedMappedExecutionStep
+]
 
 
 class _PlanBuilder:
@@ -93,12 +99,14 @@ class _PlanBuilder:
         resolved_run_config: ResolvedRunConfig,
         step_keys_to_execute: Optional[List[str]],
         known_state,
+        instance: Optional[DagsterInstance],
+        tags: Dict[str, str],
     ):
         self.pipeline = check.inst_param(pipeline, "pipeline", IPipeline)
         self.resolved_run_config = check.inst_param(
             resolved_run_config, "resolved_run_config", ResolvedRunConfig
         )
-        check.opt_list_param(step_keys_to_execute, "step_keys_to_execute", str)
+        check.opt_nullable_list_param(step_keys_to_execute, "step_keys_to_execute", str)
         self.step_keys_to_execute = step_keys_to_execute
         self.mode_definition = (
             pipeline.get_definition().get_mode_definition(resolved_run_config.mode)
@@ -110,7 +118,9 @@ class _PlanBuilder:
             SolidOutputHandle, Union[StepOutputHandle, UnresolvedStepOutputHandle]
         ] = dict()
         self.known_state = known_state
+        self._instance = instance
         self._seen_handles: Set[StepHandleUnion] = set()
+        self._tags = check.dict_param(tags, "tags", key_type=str, value_type=str)
 
     @property
     def pipeline_name(self) -> str:
@@ -128,8 +138,8 @@ class _PlanBuilder:
         self._seen_handles.add(step.handle)
         self._steps[step.solid_handle.to_string()] = step
 
-    def get_step_by_solid_handle(self, handle: SolidHandle) -> IExecutionStep:
-        check.inst_param(handle, "handle", SolidHandle)
+    def get_step_by_solid_handle(self, handle: NodeHandle) -> IExecutionStep:
+        check.inst_param(handle, "handle", NodeHandle)
         return self._steps[handle.to_string()]
 
     def get_output_handle(
@@ -148,7 +158,6 @@ class _PlanBuilder:
     def build(self) -> "ExecutionPlan":
         """Builds the execution plan"""
 
-        _check_io_manager_intermediate_storage(self.mode_definition, self.resolved_run_config)
         _check_persistent_storage_requirement(
             self.pipeline,
             self.mode_definition,
@@ -163,15 +172,17 @@ class _PlanBuilder:
         )
 
         step_dict = {step.handle: step for step in self._steps.values()}
+        step_dict_by_key = {step.key: step for step in self._steps.values()}
         step_handles_to_execute = [step.handle for step in self._steps.values()]
 
         executable_map, resolvable_map = _compute_step_maps(
             step_dict,
+            step_dict_by_key,
             step_handles_to_execute,
             self.known_state,
         )
 
-        full_plan = ExecutionPlan(
+        plan = ExecutionPlan(
             step_dict,
             executable_map,
             resolvable_map,
@@ -179,6 +190,7 @@ class _PlanBuilder:
             self.known_state,
             _compute_artifacts_persisted(
                 step_dict,
+                step_dict_by_key,
                 step_handles_to_execute,
                 pipeline_def,
                 self.resolved_run_config,
@@ -187,23 +199,32 @@ class _PlanBuilder:
         )
 
         if self.step_keys_to_execute is not None:
-            return full_plan.build_subset_plan(
+            plan = plan.build_subset_plan(
                 self.step_keys_to_execute, pipeline_def, self.resolved_run_config
             )
-        else:
-            return full_plan
+
+        # Expects that if step_keys_to_execute was set, that the `plan` variable will have the
+        # reflected step_keys_to_execute
+        if pipeline_def.is_using_memoization(self._tags):
+            if self.step_keys_to_execute is not None:
+                raise DagsterInvariantViolationError(
+                    "Cannot use both memoization and re-execution at this time."
+                )
+            plan = plan.build_memoized_plan(pipeline_def, self.resolved_run_config, self._instance)
+
+        return plan
 
     def _build_from_sorted_solids(
         self,
-        solids: List[Solid],
+        solids: List[Node],
         dependency_structure: DependencyStructure,
-        parent_handle: Optional[SolidHandle] = None,
+        parent_handle: Optional[NodeHandle] = None,
         parent_step_inputs: Optional[
             List[Union[StepInput, UnresolvedMappedStepInput, UnresolvedCollectStepInput]]
         ] = None,
     ):
         for solid in solids:
-            handle = SolidHandle(solid.name, parent_handle)
+            handle = NodeHandle(solid.name, parent_handle)
 
             ### 1. INPUTS
             # Create and add execution plan steps for solid inputs
@@ -346,21 +367,21 @@ class _PlanBuilder:
 
 def get_step_input_source(
     plan_builder: _PlanBuilder,
-    solid: Solid,
+    solid: Node,
     input_name: str,
     input_def: InputDefinition,
     dependency_structure: DependencyStructure,
-    handle: SolidHandle,
+    handle: NodeHandle,
     parent_step_inputs: Optional[
         List[Union[StepInput, UnresolvedMappedStepInput, UnresolvedCollectStepInput]]
     ],
 ):
     check.inst_param(plan_builder, "plan_builder", _PlanBuilder)
-    check.inst_param(solid, "solid", Solid)
+    check.inst_param(solid, "solid", Node)
     check.str_param(input_name, "input_name")
     check.inst_param(input_def, "input_def", InputDefinition)
     check.inst_param(dependency_structure, "dependency_structure", DependencyStructure)
-    check.opt_inst_param(handle, "handle", SolidHandle)
+    check.opt_inst_param(handle, "handle", NodeHandle)
     check.opt_list_param(
         parent_step_inputs,
         "parent_step_inputs",
@@ -497,11 +518,13 @@ def get_step_input_source(
     # Otherwise we throw an error.
     raise DagsterInvariantViolationError(
         (
-            "In pipeline {pipeline_name} solid {solid_name}, input {input_name} "
+            "In {described_target} {described_node}, input {input_name} "
             "must get a value either (a) from a dependency or (b) from the "
             "inputs section of its configuration."
         ).format(
-            pipeline_name=plan_builder.pipeline_name, solid_name=solid.name, input_name=input_name
+            described_target=plan_builder.pipeline.get_definition().describe_target(),
+            described_node=solid.describe_node(),
+            input_name=input_name,
         )
     )
 
@@ -516,6 +539,8 @@ class ExecutionPlan(
             ("step_handles_to_execute", List[StepHandleUnion]),
             ("known_state", KnownExecutionState),
             ("artifacts_persisted", bool),
+            ("step_output_versions", Dict[StepOutputHandle, str]),
+            ("step_dict_by_key", Dict[str, IExecutionStep]),
         ],
     )
 ):
@@ -527,6 +552,8 @@ class ExecutionPlan(
         step_handles_to_execute,
         known_state=None,
         artifacts_persisted=False,
+        step_output_versions=None,
+        step_dict_by_key=None,
     ):
         return super(ExecutionPlan, cls).__new__(
             cls,
@@ -549,6 +576,24 @@ class ExecutionPlan(
             ),
             known_state=check.opt_inst_param(known_state, "known_state", KnownExecutionState),
             artifacts_persisted=check.bool_param(artifacts_persisted, "artifacts_persisted"),
+            step_output_versions=check.opt_dict_param(
+                step_output_versions,
+                "step_output_versions",
+                key_type=StepOutputHandle,
+                value_type=str,
+            ),
+            step_dict_by_key={step.key: step for step in step_dict.values()}
+            if step_dict_by_key is None
+            else check.dict_param(
+                step_dict_by_key,
+                "step_dict_by_key",
+                key_type=str,
+                value_type=(
+                    ExecutionStep,
+                    UnresolvedMappedExecutionStep,
+                    UnresolvedCollectExecutionStep,
+                ),
+            ),
         )
 
     @property
@@ -561,14 +606,14 @@ class ExecutionPlan(
 
     def get_step_output(self, step_output_handle: StepOutputHandle) -> StepOutput:
         check.inst_param(step_output_handle, "step_output_handle", StepOutputHandle)
-        return _get_step_output(self.step_dict, step_output_handle)
+        return _get_step_output(self.step_dict_by_key, step_output_handle)
 
     def get_manager_key(
         self,
         step_output_handle: StepOutputHandle,
         pipeline_def: PipelineDefinition,
     ) -> str:
-        return _get_manager_key(self.step_dict, step_output_handle, pipeline_def)
+        return _get_manager_key(self.step_dict_by_key, step_output_handle, pipeline_def)
 
     def has_step(self, handle: StepHandleUnion) -> bool:
         check.inst_param(handle, "handle", StepHandleTypes)
@@ -579,7 +624,10 @@ class ExecutionPlan(
         return self.step_dict[handle]
 
     def get_step_by_key(self, key: str) -> IExecutionStep:
-        return _get_step_by_key(self.step_dict, key)
+        check.str_param(key, "key")
+        if not key in self.step_dict_by_key:
+            check.failed(f"plan has no step with key {key}")
+        return self.step_dict_by_key[key]
 
     def get_executable_step_by_key(self, key: str) -> ExecutionStep:
         step = self.get_step_by_key(key)
@@ -611,7 +659,7 @@ class ExecutionPlan(
 
     def get_steps_to_execute_by_level(self) -> List[List[ExecutionStep]]:
         return _get_steps_to_execute_by_level(
-            self.step_dict, self.step_handles_to_execute, self.executable_map
+            self.step_dict, self.step_dict_by_key, self.step_handles_to_execute, self.executable_map
         )
 
     def get_executable_step_deps(self) -> Dict[str, Set[str]]:
@@ -629,6 +677,7 @@ class ExecutionPlan(
 
         _update_from_resolved_dynamic_outputs(
             self.step_dict,
+            self.step_dict_by_key,
             self.executable_map,
             self.resolvable_map,
             self.step_handles_to_execute,
@@ -644,6 +693,7 @@ class ExecutionPlan(
         step_keys_to_execute: List[str],
         pipeline_def: PipelineDefinition,
         resolved_run_config: ResolvedRunConfig,
+        step_output_versions=None,
     ) -> "ExecutionPlan":
         check.list_param(step_keys_to_execute, "step_keys_to_execute", of_type=str)
         step_handles_to_execute = [StepHandle.parse_from_key(key) for key in step_keys_to_execute]
@@ -661,6 +711,7 @@ class ExecutionPlan(
 
         executable_map, resolvable_map = _compute_step_maps(
             self.step_dict,
+            self.step_dict_by_key,
             step_handles_to_execute,
             self.known_state,
         )
@@ -673,11 +724,114 @@ class ExecutionPlan(
             self.known_state,
             _compute_artifacts_persisted(
                 self.step_dict,
+                self.step_dict_by_key,
                 step_handles_to_execute,
                 pipeline_def,
                 resolved_run_config,
                 executable_map,
             ),
+            step_output_versions=step_output_versions,
+        )
+
+    def get_version_for_step_output_handle(
+        self, step_output_handle: StepOutputHandle
+    ) -> Optional[str]:
+        return self.step_output_versions.get(step_output_handle)
+
+    def build_memoized_plan(
+        self,
+        pipeline_def: PipelineDefinition,
+        resolved_run_config: ResolvedRunConfig,
+        instance: Optional[DagsterInstance],
+    ) -> "ExecutionPlan":
+        """
+        Returns:
+            ExecutionPlan: Execution plan that runs only unmemoized steps.
+        """
+        from ..build_resources import build_resources, initialize_console_manager
+        from ...storage.memoizable_io_manager import MemoizableIOManager
+        from ..resources_init import get_dependencies, resolve_resource_dependencies
+
+        instance = check.opt_inst_param(
+            instance, "instance", DagsterInstance, default=DagsterInstance.get()
+        )
+        mode = resolved_run_config.mode
+        mode_def = pipeline_def.get_mode_definition(mode)
+
+        # Memoization cannot be used with dynamic orchestration yet.
+        # Tracking: https://github.com/dagster-io/dagster/issues/4451
+        for node_def in pipeline_def.all_node_defs:
+            if pipeline_def.dependency_structure.is_dynamic_mapped(
+                node_def.name
+            ) or pipeline_def.dependency_structure.has_dynamic_downstreams(node_def.name):
+                raise DagsterInvariantViolationError(
+                    "Attempted to use memoization with dynamic orchestration, which is not yet "
+                    "supported."
+                )
+
+        unmemoized_step_keys = set()
+
+        log_manager = initialize_console_manager(None)
+
+        step_output_versions = resolve_step_output_versions(pipeline_def, self, resolved_run_config)
+
+        resource_defs_to_init = {}
+        io_manager_keys = {}  # Map step output handles to io manager keys
+
+        for step in self.steps:
+            for output_name in cast(ExecutionStepUnion, step).step_output_dict.keys():
+                step_output_handle = StepOutputHandle(step.key, output_name)
+
+                io_manager_key = self.get_manager_key(step_output_handle, pipeline_def)
+                io_manager_keys[step_output_handle] = io_manager_key
+
+                resource_deps = resolve_resource_dependencies(mode_def.resource_defs)
+                resource_keys_to_init = get_dependencies(io_manager_key, resource_deps)
+                for resource_key in resource_keys_to_init:
+                    resource_defs_to_init[resource_key] = mode_def.resource_defs[resource_key]
+
+        all_resources_config = resolved_run_config.to_dict().get("resources", {})
+        resource_config = {
+            resource_key: config_val
+            for resource_key, config_val in all_resources_config.items()
+            if resource_key in resource_defs_to_init
+        }
+
+        with build_resources(
+            resources=resource_defs_to_init,
+            instance=instance,
+            resource_config=resource_config,
+            log_manager=log_manager,
+        ) as resources:
+            for step_output_handle, io_manager_key in io_manager_keys.items():
+                io_manager = getattr(resources, io_manager_key)
+                if not isinstance(io_manager, MemoizableIOManager):
+                    raise DagsterInvariantViolationError(
+                        f"{pipeline_def.describe_target().capitalize()} uses memoization, but IO manager "
+                        f"'{io_manager_key}' is not a MemoizableIOManager. In order to use "
+                        "memoization, all io managers need to subclass MemoizableIOManager. "
+                        "Learn more about MemoizableIOManagers here: "
+                        "https://docs.dagster.io/_apidocs/internals#memoizable-io-manager-experimental."
+                    )
+                context = get_output_context(
+                    execution_plan=self,
+                    pipeline_def=pipeline_def,
+                    resolved_run_config=resolved_run_config,
+                    step_output_handle=step_output_handle,
+                    run_id=None,
+                    log_manager=log_manager,
+                    step_context=None,
+                    resources=resources,
+                    version=step_output_versions[step_output_handle],
+                )
+                if not io_manager.has_output(context):
+                    unmemoized_step_keys.add(step_output_handle.step_key)
+
+        return self.build_subset_plan(
+            list(unmemoized_step_keys),
+            pipeline_def,
+            resolved_run_config,
+            step_output_versions=step_output_versions,
         )
 
     def start(
@@ -718,6 +872,8 @@ class ExecutionPlan(
         resolved_run_config: ResolvedRunConfig,
         step_keys_to_execute: Optional[List[str]] = None,
         known_state=None,
+        instance=None,
+        tags=None,
     ) -> "ExecutionPlan":
         """Here we build a new ExecutionPlan from a pipeline definition and the resolved run config.
 
@@ -729,14 +885,17 @@ class ExecutionPlan(
         """
         check.inst_param(pipeline, "pipeline", IPipeline)
         check.inst_param(resolved_run_config, "resolved_run_config", ResolvedRunConfig)
-        check.opt_list_param(step_keys_to_execute, "step_keys_to_execute", of_type=str)
+        check.opt_nullable_list_param(step_keys_to_execute, "step_keys_to_execute", of_type=str)
         check.opt_inst_param(known_state, "known_state", KnownExecutionState)
+        tags = check.opt_dict_param(tags, "tags", key_type=str, value_type=str)
 
         plan_builder = _PlanBuilder(
             pipeline,
             resolved_run_config=resolved_run_config,
             step_keys_to_execute=step_keys_to_execute,
             known_state=known_state,
+            instance=instance,
+            tags=tags,
         )
 
         # Finally, we build and return the execution plan
@@ -782,6 +941,7 @@ class ExecutionPlan(
             )
 
         step_dict = {}
+        step_dict_by_key = {}
 
         for step_snap in execution_plan_snapshot.steps:
             input_snaps = step_snap.inputs
@@ -829,6 +989,7 @@ class ExecutionPlan(
                 raise Exception(f"Unexpected step kind {str(step_snap.kind)}")
 
             step_dict[step.handle] = step
+            step_dict_by_key[step.key] = step
 
         step_handles_to_execute = [
             StepHandle.parse_from_key(key) for key in execution_plan_snapshot.step_keys_to_execute
@@ -836,10 +997,16 @@ class ExecutionPlan(
 
         executable_map, resolvable_map = _compute_step_maps(
             step_dict,
+            step_dict_by_key,
             step_handles_to_execute,
             execution_plan_snapshot.initial_known_state,
         )
 
+        step_output_versions: Dict[StepOutputHandle, str] = {}
+        for step_output_version_data in execution_plan_snapshot.step_output_versions:
+            step_output_handle = step_output_version_data.step_output_handle
+            step_output_versions[step_output_handle] = step_output_version_data.version
+
         return ExecutionPlan(
             step_dict,
             executable_map,
@@ -847,11 +1014,13 @@ class ExecutionPlan(
             step_handles_to_execute,
             execution_plan_snapshot.initial_known_state,
             execution_plan_snapshot.artifacts_persisted,
+            step_output_versions,
         )
 
 
 def _update_from_resolved_dynamic_outputs(
     step_dict: Dict[StepHandleUnion, IExecutionStep],
+    step_dict_by_key: Dict[str, IExecutionStep],
     executable_map: Dict[str, Union[StepHandle, ResolvedFromDynamicStepHandle]],
     resolvable_map: Dict[FrozenSet[str], List[UnresolvedStepHandle]],
     step_handles_to_execute: List[StepHandleUnion],
@@ -883,6 +1052,7 @@ def _update_from_resolved_dynamic_outputs(
     for step in resolved_steps:
         step_dict[step.handle] = step
         executable_map[step.key] = step.handle
+        step_dict_by_key[step.key] = step
 
     for key_set in key_sets_to_clear:
         del resolvable_map[key_set]
@@ -899,7 +1069,7 @@ def can_isolate_steps(pipeline_def: PipelineDefinition, mode_def: ModeDefinition
 
     output_defs = [
         output_def
-        for solid_def in pipeline_def.all_solid_defs
+        for solid_def in pipeline_def.all_node_defs
         for output_def in solid_def.output_defs
     ]
     for output_def in output_defs:
@@ -918,48 +1088,28 @@ def _check_persistent_storage_requirement(
 
     pipeline_def = pipeline.get_definition()
     executor_def = executor_def_from_config(mode_def, resolved_run_config)
-    if ExecutorRequirement.PERSISTENT_OUTPUTS not in executor_def.requirements:
+    requirements_lst = executor_def.get_requirements(
+        resolved_run_config.execution.execution_engine_config
+    )
+    if ExecutorRequirement.PERSISTENT_OUTPUTS not in requirements_lst:
         return
 
-    intermediate_storage_def = resolved_run_config.intermediate_storage_def_for_mode(mode_def)
-
-    if not (
-        can_isolate_steps(pipeline_def, mode_def)
-        or (intermediate_storage_def and intermediate_storage_def.is_persistent)
-    ):
-        raise DagsterUnmetExecutorRequirementsError(
-            "You have attempted to use an executor that uses multiple processes, but your pipeline "
-            "includes solid outputs that will not be stored somewhere where other processes can "
-            "retrieve them. Please use a persistent IO manager for these outputs. E.g. with\n"
-            '    @pipeline(mode_defs=[ModeDefinition(resource_defs={"io_manager": fs_io_manager})])'
-        )
-
-
-def _check_io_manager_intermediate_storage(
-    mode_def: ModeDefinition, resolved_run_config: ResolvedRunConfig
-) -> None:
-    """Only one of io_manager and intermediate_storage should be set."""
-    # pylint: disable=comparison-with-callable
-    from dagster.core.storage.system_storage import mem_intermediate_storage
-
-    intermediate_storage_def = resolved_run_config.intermediate_storage_def_for_mode(mode_def)
-    intermediate_storage_is_default = (
-        intermediate_storage_def is None or intermediate_storage_def == mem_intermediate_storage
-    )
-
-    io_manager = mode_def.resource_defs["io_manager"]
-    io_manager_is_default = io_manager == mem_io_manager
-
-    if not intermediate_storage_is_default and not io_manager_is_default:
-        raise DagsterInvariantViolationError(
-            'You have specified an intermediate storage, "{intermediate_storage_name}", and have '
-            "also specified a default IO manager. You must specify only one. To avoid specifying "
-            "an intermediate storage, omit the intermediate_storage_defs argument to your "
-            'ModeDefinition and omit "intermediate_storage" in your run config. To avoid '
-            'specifying a default IO manager, omit the "io_manager" key from the '
-            "resource_defs argument to your ModeDefinition.".format(
-                intermediate_storage_name=intermediate_storage_def.name
+    if not can_isolate_steps(pipeline_def, mode_def):
+        if isinstance(pipeline_def, JobDefinition):
+            target = "job"
+            node = "op"
+            suggestion = 'the_graph.to_job(resource_defs={"io_manager": fs_io_manager})'
+        else:
+            target = "pipeline"
+            node = "solid"
+            suggestion = (
+                '@pipeline(mode_defs=[ModeDefinition(resource_defs={"io_manager": fs_io_manager})])'
             )
+        raise DagsterUnmetExecutorRequirementsError(
+            f"You have attempted to use an executor that uses multiple processes, but your {target} "
+            f"includes {node} outputs that will not be stored somewhere where other processes can "
+            "retrieve them. Please use a persistent IO manager for these outputs. E.g. with\n"
+            f"    {suggestion}"
         )
 
 
@@ -1020,7 +1170,12 @@ def should_skip_step(execution_plan: ExecutionPlan, instance: DagsterInstance, r
 
 
 def _compute_artifacts_persisted(
-    step_dict, step_handles_to_execute, pipeline_def, resolved_run_config, executable_map
+    step_dict,
+    step_dict_by_key,
+    step_handles_to_execute,
+    pipeline_def,
+    resolved_run_config,
+    executable_map,
 ):
     """
     Check if all the border steps of the current run have non-in-memory IO managers for reexecution.
@@ -1029,21 +1184,13 @@ def _compute_artifacts_persisted(
     """
     # pylint: disable=comparison-with-callable
 
-    # intermediate storage backcomopat
-    # https://github.com/dagster-io/dagster/issues/3043
-
     mode_def = pipeline_def.get_mode_definition(resolved_run_config.mode)
 
-    if mode_def.get_intermediate_storage_def(
-        resolved_run_config.intermediate_storage.intermediate_storage_name
-    ).is_persistent:
-        return True
-
     if len(step_dict) == 0:
         return False
 
     steps_by_level = _get_steps_to_execute_by_level(
-        step_dict, step_handles_to_execute, executable_map
+        step_dict, step_dict_by_key, step_handles_to_execute, executable_map
     )
 
     if len(steps_by_level) == 0:
@@ -1053,7 +1200,9 @@ def _compute_artifacts_persisted(
         # check if all its inputs' upstream step outputs have non-in-memory IO manager configured
         for step_input in step.step_inputs:
             for step_output_handle in step_input.get_step_output_handle_dependencies():
-                io_manager_key = _get_manager_key(step_dict, step_output_handle, pipeline_def)
+                io_manager_key = _get_manager_key(
+                    step_dict_by_key, step_output_handle, pipeline_def
+                )
                 manager_def = mode_def.resource_defs.get(io_manager_key)
                 if (
                     # no IO manager is configured
@@ -1073,12 +1222,11 @@ def _get_step_by_key(step_dict, key) -> IExecutionStep:
     check.failed(f"plan has no step with key {key}")
 
 
-def _get_steps_to_execute_by_level(step_dict, step_handles_to_execute, executable_map):
+def _get_steps_to_execute_by_level(
+    step_dict, step_dict_by_key, step_handles_to_execute, executable_map
+):
     return [
-        [
-            cast(ExecutionStep, _get_step_by_key(step_dict, step_key))
-            for step_key in sorted(step_key_level)
-        ]
+        [cast(ExecutionStep, step_dict_by_key[step_key]) for step_key in sorted(step_key_level)]
         for step_key_level in toposort(
             _get_executable_step_deps(step_dict, step_handles_to_execute, executable_map)
         )
@@ -1118,14 +1266,14 @@ def _get_executable_step_deps(
     return deps
 
 
-def _get_step_output(step_dict, step_output_handle: StepOutputHandle) -> StepOutput:
+def _get_step_output(step_dict_by_key, step_output_handle: StepOutputHandle) -> StepOutput:
     check.inst_param(step_output_handle, "step_output_handle", StepOutputHandle)
-    step = _get_step_by_key(step_dict, step_output_handle.step_key)
+    step = step_dict_by_key[step_output_handle.step_key]
     return step.step_output_named(step_output_handle.output_name)
 
 
-def _get_manager_key(step_dict, step_output_handle, pipeline_def):
-    step_output = _get_step_output(step_dict, step_output_handle)
+def _get_manager_key(step_dict_by_key, step_output_handle, pipeline_def):
+    step_output = _get_step_output(step_dict_by_key, step_output_handle)
     solid_handle = step_output.solid_handle
     output_def = pipeline_def.get_solid(solid_handle).output_def_named(step_output.name)
     return output_def.io_manager_key
@@ -1133,7 +1281,7 @@ def _get_manager_key(step_dict, step_output_handle, pipeline_def):
 
 # computes executable_map and resolvable_map and returns them as a tuple. Also
 # may modify the passed in step_dict to include resolved step using known_state.
-def _compute_step_maps(step_dict, step_handles_to_execute, known_state):
+def _compute_step_maps(step_dict, step_dict_by_key, step_handles_to_execute, known_state):
     check.list_param(
         step_handles_to_execute,
         "step_handles_to_execute",
@@ -1165,7 +1313,7 @@ def _compute_step_maps(step_dict, step_handles_to_execute, known_state):
             for key in step.resolved_by_step_keys:
                 if key not in step_keys_to_execute:
                     raise DagsterInvariantViolationError(
-                        f'Unresolved ExecutionStep "{step.key}" is resolved by "{step.resolved_by_step_key}" '
+                        f'Unresolved ExecutionStep "{step.key}" is resolved by "{key}" '
                         "which is not part of the current step selection"
                     )
 
@@ -1178,6 +1326,7 @@ def _compute_step_maps(step_dict, step_handles_to_execute, known_state):
     if known_state:
         _update_from_resolved_dynamic_outputs(
             step_dict,
+            step_dict_by_key,
             executable_map,
             resolvable_map,
             step_handles_to_execute,
