@@ -1,5 +1,6 @@
 import inspect
 import logging
+import logging.config
 import os
 import sys
 import tempfile
@@ -8,7 +9,19 @@ import warnings
 import weakref
 from collections import defaultdict
 from enum import Enum
-from typing import TYPE_CHECKING, Any, Callable, Dict, Iterable, List, Optional, Tuple, Union, cast
+from typing import (
+    TYPE_CHECKING,
+    Any,
+    Callable,
+    Dict,
+    Iterable,
+    List,
+    Optional,
+    Set,
+    Tuple,
+    Union,
+    cast,
+)
 
 import yaml
 from dagster import check
@@ -18,11 +31,11 @@ from dagster.core.definitions.pipeline_base import InMemoryPipeline
 from dagster.core.errors import (
     DagsterHomeNotSetError,
     DagsterInvariantViolationError,
-    DagsterNoStepsToExecuteException,
     DagsterRunAlreadyExists,
     DagsterRunConflict,
 )
 from dagster.core.storage.pipeline_run import (
+    IN_PROGRESS_RUN_STATUSES,
     PipelineRun,
     PipelineRunStatsSnapshot,
     PipelineRunStatus,
@@ -34,6 +47,7 @@ from dagster.core.system_config.objects import ResolvedRunConfig
 from dagster.core.utils import str_format_list
 from dagster.serdes import ConfigurableClass
 from dagster.seven import get_current_datetime_in_utc
+from dagster.utils.backcompat import experimental_functionality_warning
 from dagster.utils.error import serializable_error_info_from_exc_info
 
 from .config import DAGSTER_CONFIG_YAML_FILENAME, is_dagster_home_set
@@ -67,10 +81,6 @@ if TYPE_CHECKING:
     from dagster.core.debug import DebugRunPayload
 
 
-def is_memoized_run(tags: Dict[str, Any]) -> bool:
-    return tags is not None and MEMOIZED_RUN_TAG in tags and tags.get(MEMOIZED_RUN_TAG) == "true"
-
-
 def _check_run_equality(
     pipeline_run: PipelineRun, candidate_run: PipelineRun
 ) -> Dict[str, Tuple[Any, Any]]:
@@ -178,7 +188,7 @@ class DagsterInstance:
     For example, to use Postgres for run and event log storage, you can write a ``dagster.yaml``
     such as the following:
 
-    .. literalinclude:: ../../../../docs/sections/deploying/postgres_dagster.yaml
+    .. literalinclude:: ../../../../../examples/docs_snippets/docs_snippets/deploying/postgres_dagster.yaml
        :caption: dagster.yaml
        :language: YAML
 
@@ -225,10 +235,10 @@ class DagsterInstance:
         run_storage: "RunStorage",
         event_storage: "EventLogStorage",
         compute_log_manager: "ComputeLogManager",
-        schedule_storage: Optional["ScheduleStorage"] = None,
+        run_coordinator: "RunCoordinator",
+        run_launcher: "RunLauncher",
         scheduler: Optional["Scheduler"] = None,
-        run_coordinator: Optional["RunCoordinator"] = None,
-        run_launcher: Optional["RunLauncher"] = None,
+        schedule_storage: Optional["ScheduleStorage"] = None,
         settings: Optional[Dict[str, Any]] = None,
         ref: Optional[InstanceRef] = None,
     ):
@@ -299,6 +309,12 @@ class DagsterInstance:
 
         self._subscribers: Dict[str, List[Callable]] = defaultdict(list)
 
+        if self.run_monitoring_enabled:
+            check.invariant(
+                self.run_launcher.supports_check_run_worker_health,
+                "Run monitoring only supports select RunLaunchers",
+            )
+
     # ctors
 
     @staticmethod
@@ -316,7 +332,7 @@ class DagsterInstance:
             tempdir = DagsterInstance.temp_storage()
 
         return DagsterInstance(
-            InstanceType.EPHEMERAL,
+            instance_type=InstanceType.EPHEMERAL,
             local_artifact_storage=LocalArtifactStorage(tempdir),
             run_storage=InMemoryRunStorage(preload=preload),
             event_storage=InMemoryEventLogStorage(preload=preload),
@@ -489,11 +505,11 @@ class DagsterInstance:
     # schedule storage
 
     @property
-    def schedule_storage(self) -> "ScheduleStorage":
+    def schedule_storage(self) -> Optional["ScheduleStorage"]:
         return self._schedule_storage
 
     @property
-    def scheduler(self) -> "Scheduler":
+    def scheduler(self) -> Optional["Scheduler"]:
         return self._scheduler
 
     @property
@@ -541,6 +557,54 @@ class DagsterInstance:
         else:
             return dagster_telemetry_enabled_default
 
+    # run monitoring
+
+    @property
+    def run_monitoring_enabled(self) -> bool:
+        if self.is_ephemeral:
+            return False
+
+        run_monitoring_enabled_default = False
+
+        run_monitoring_settings = self.get_settings("run_monitoring")
+
+        if not run_monitoring_settings:
+            return run_monitoring_enabled_default
+
+        if "enabled" in run_monitoring_settings:
+            return run_monitoring_settings["enabled"]
+        else:
+            return run_monitoring_enabled_default
+
+    @property
+    def run_monitoring_settings(self) -> Dict:
+        check.invariant(self.run_monitoring_enabled, "run_monitoring is not enabled")
+        return self.get_settings("run_monitoring")
+
+    @property
+    def run_monitoring_start_timeout_seconds(self) -> int:
+        return self.run_monitoring_settings.get("start_timeout_seconds", 180)
+
+    @property
+    def run_monitoring_max_resume_run_attempts(self) -> int:
+        return self.run_monitoring_settings.get("max_resume_run_attempts", 3)
+
+    @property
+    def run_monitoring_poll_interval_seconds(self) -> int:
+        return self.run_monitoring_settings.get("poll_interval_seconds", 120)
+
+    # python logs
+
+    @property
+    def managed_python_loggers(self) -> List[str]:
+        python_log_settings = self.get_settings("python_logs") or {}
+        return python_log_settings.get("managed_python_loggers", [])
+
+    @property
+    def python_log_level(self) -> Optional[str]:
+        python_log_settings = self.get_settings("python_logs") or {}
+        return python_log_settings.get("python_log_level")
+
     def upgrade(self, print_fn=None):
         from dagster.core.storage.migration.utils import upgrading_instance
 
@@ -562,7 +626,6 @@ class DagsterInstance:
 
     def optimize_for_dagit(self, statement_timeout):
         if self._schedule_storage:
-            self._schedule_storage.validate_stored_schedules(self.scheduler_class)
             self._schedule_storage.optimize_for_dagit(statement_timeout=statement_timeout)
         self._run_storage.optimize_for_dagit(statement_timeout=statement_timeout)
         self._event_storage.optimize_for_dagit(statement_timeout=statement_timeout)
@@ -582,7 +645,7 @@ class DagsterInstance:
 
     # run storage
 
-    def get_run_by_id(self, run_id: str) -> PipelineRun:
+    def get_run_by_id(self, run_id: str) -> Optional[PipelineRun]:
         return self._run_storage.get_run_by_id(run_id)
 
     def get_pipeline_snapshot(self, snapshot_id: str) -> "PipelineSnapshot":
@@ -616,10 +679,10 @@ class DagsterInstance:
     def get_run_step_stats(self, run_id, step_keys=None) -> List["RunStepKeyStatsSnapshot"]:
         return self._event_storage.get_step_stats_for_run(run_id, step_keys)
 
-    def get_run_tags(self) -> Dict[str, Any]:
+    def get_run_tags(self) -> List[Tuple[str, Set[str]]]:
         return self._run_storage.get_run_tags()
 
-    def get_run_group(self, run_id: str) -> Tuple[str, List[PipelineRun]]:
+    def get_run_group(self, run_id: str) -> Optional[Tuple[str, Iterable[PipelineRun]]]:
         return self._run_storage.get_run_group(run_id)
 
     def create_run_for_pipeline(
@@ -630,7 +693,6 @@ class DagsterInstance:
         run_config=None,
         mode=None,
         solids_to_execute=None,
-        step_keys_to_execute=None,
         status=None,
         tags=None,
         root_run_id=None,
@@ -670,49 +732,19 @@ class DagsterInstance:
                     solids_to_execute=solids_to_execute
                 )
 
+        step_keys_to_execute = None
+
         if execution_plan:
-            final_execution_plan = execution_plan
-            check.invariant(
-                step_keys_to_execute is None,
-                "Should not pass execution_plan and step_keys_to_execute to create_run",
-            )
             step_keys_to_execute = execution_plan.step_keys_to_execute
+
         else:
             resolved_run_config = ResolvedRunConfig.build(pipeline_def, run_config, mode)
-            full_execution_plan = ExecutionPlan.build(
+            execution_plan = ExecutionPlan.build(
                 InMemoryPipeline(pipeline_def),
                 resolved_run_config,
+                tags=tags,
             )
 
-            if is_memoized_run(tags):
-                from dagster.core.execution.resolve_versions import resolve_memoized_execution_plan
-
-                if step_keys_to_execute:
-                    raise DagsterInvariantViolationError(
-                        "step_keys_to_execute parameter cannot be used in conjunction with memoized "
-                        "pipeline runs."
-                    )
-
-                final_execution_plan = resolve_memoized_execution_plan(
-                    full_execution_plan,
-                    pipeline_def,
-                    run_config,
-                    self,
-                    resolved_run_config,
-                )  # TODO: tighter integration with existing step_keys_to_execute functionality
-                step_keys_to_execute = final_execution_plan.step_keys_to_execute
-                if not step_keys_to_execute:
-                    raise DagsterNoStepsToExecuteException(
-                        "No steps found to execute. "
-                        "This is because every step in the plan has already been memoized."
-                    )
-            elif step_keys_to_execute:
-                final_execution_plan = full_execution_plan.build_subset_plan(
-                    step_keys_to_execute, pipeline_def, resolved_run_config
-                )
-            else:
-                final_execution_plan = full_execution_plan
-
         return self.create_run(
             pipeline_name=pipeline_def.name,
             run_id=run_id,
@@ -727,7 +759,7 @@ class DagsterInstance:
             parent_run_id=parent_run_id,
             pipeline_snapshot=pipeline_def.get_pipeline_snapshot(),
             execution_plan_snapshot=snapshot_from_execution_plan(
-                final_execution_plan,
+                execution_plan,
                 pipeline_def.get_pipeline_snapshot_id(),
             ),
             parent_pipeline_snapshot=pipeline_def.get_parent_pipeline_snapshot(),
@@ -842,7 +874,7 @@ class DagsterInstance:
 
         check.inst_param(execution_plan_snapshot, "execution_plan_snapshot", ExecutionPlanSnapshot)
         check.str_param(pipeline_snapshot_id, "pipeline_snapshot_id")
-        check.opt_list_param(step_keys_to_execute, "step_keys_to_execute", of_type=str)
+        check.opt_nullable_list_param(step_keys_to_execute, "step_keys_to_execute", of_type=str)
 
         check.invariant(
             execution_plan_snapshot.pipeline_snapshot_id == pipeline_snapshot_id,
@@ -856,17 +888,6 @@ class DagsterInstance:
             ),
         )
 
-        check.invariant(
-            set(step_keys_to_execute) == set(execution_plan_snapshot.step_keys_to_execute)
-            if step_keys_to_execute
-            else set(execution_plan_snapshot.step_keys_to_execute)
-            == set([step.key for step in execution_plan_snapshot.steps]),
-            "We encode step_keys_to_execute twice in our stack, unfortunately. This check "
-            "ensures that they are consistent. We check that step_keys_to_execute in the plan "
-            "matches the step_keys_to_execute params if it is set. If it is not, this indicates "
-            "a full execution plan, and so we verify that.",
-        )
-
         execution_plan_snapshot_id = create_execution_plan_snapshot_id(execution_plan_snapshot)
 
         if not self._run_storage.has_execution_plan_snapshot(execution_plan_snapshot_id):
@@ -1041,8 +1062,19 @@ class DagsterInstance:
 
     # event storage
 
-    def logs_after(self, run_id, cursor, of_type: "DagsterEventType" = None):
-        return self._event_storage.get_logs_for_run(run_id, cursor=cursor, of_type=of_type)
+    def logs_after(
+        self,
+        run_id,
+        cursor,
+        of_type: "DagsterEventType" = None,
+        limit: Optional[int] = None,
+    ):
+        return self._event_storage.get_logs_for_run(
+            run_id,
+            cursor=cursor,
+            of_type=of_type,
+            limit=limit,
+        )
 
     def all_logs(self, run_id, of_type: "DagsterEventType" = None):
         return self._event_storage.get_logs_for_run(run_id, of_type=of_type)
@@ -1058,6 +1090,9 @@ class DagsterInstance:
     def all_asset_keys(self):
         return self._event_storage.all_asset_keys()
 
+    def get_asset_keys(self, prefix=None, limit=None, cursor=None):
+        return self._event_storage.get_asset_keys(prefix=prefix, limit=limit, cursor=cursor)
+
     def has_asset_key(self, asset_key: AssetKey) -> bool:
         return self._event_storage.has_asset_key(asset_key)
 
@@ -1142,11 +1177,42 @@ records = instance.get_event_records(
 
     # event subscriptions
 
-    def get_logger(self):
-        logger = logging.Logger("__event_listener")
-        logger.addHandler(_EventListenerLogHandler(self))
-        logger.setLevel(10)
-        return logger
+    def _get_yaml_python_handlers(self):
+        if self._settings:
+            logging_config = self.get_settings("python_logs").get("dagster_handler_config", {})
+
+            if logging_config:
+                experimental_functionality_warning("Handling yaml-defined logging configuration")
+
+            # Handlers can only be retrieved from dictConfig configuration if they are attached
+            # to a logger. We add a dummy logger to the configuration that allows us to access user
+            # defined handlers.
+            handler_names = logging_config.get("handlers", {}).keys()
+
+            dagster_dummy_logger_name = "dagster_dummy_logger"
+
+            processed_dict_conf = {
+                "version": 1,
+                "disable_existing_loggers": False,
+                "loggers": {dagster_dummy_logger_name: {"handlers": handler_names}},
+            }
+            processed_dict_conf.update(logging_config)
+
+            logging.config.dictConfig(processed_dict_conf)
+
+            dummy_logger = logging.getLogger(dagster_dummy_logger_name)
+            return dummy_logger.handlers
+        return []
+
+    def _get_event_log_handler(self):
+        event_log_handler = _EventListenerLogHandler(self)
+        event_log_handler.setLevel(10)
+        return event_log_handler
+
+    def get_handlers(self):
+        handlers = [self._get_event_log_handler()]
+        handlers.extend(self._get_yaml_python_handlers())
+        return handlers
 
     def handle_new_event(self, event):
         run_id = event.run_id
@@ -1198,6 +1264,7 @@ records = instance.get_event_records(
             pipeline_name=pipeline_run.pipeline_name,
             message=message,
             event_specific_data=engine_event_data,
+            step_key=step_key,
         )
         event_record = EventLogEntry(
             message=message,
@@ -1223,7 +1290,7 @@ records = instance.get_event_records(
         message = check.opt_str_param(
             message,
             "message",
-            "Sending pipeline termination request.",
+            "Sending run termination request.",
         )
         canceling_event = DagsterEvent(
             event_type_value=DagsterEventType.PIPELINE_CANCELING.value,
@@ -1257,7 +1324,7 @@ records = instance.get_event_records(
         message = check.opt_str_param(
             message,
             "mesage",
-            "This pipeline run has been marked as canceled from outside the execution context.",
+            "This run has been marked as canceled from outside the execution context.",
         )
 
         dagster_event = DagsterEvent(
@@ -1288,7 +1355,7 @@ records = instance.get_event_records(
         message = check.opt_str_param(
             message,
             "message",
-            "This pipeline run has been marked as failed from outside the execution context.",
+            "This run has been marked as failed from outside the execution context.",
         )
 
         dagster_event = DagsterEvent(
@@ -1315,9 +1382,6 @@ records = instance.get_event_records(
     def file_manager_directory(self, run_id):
         return self._local_artifact_storage.file_manager_dir(run_id)
 
-    def intermediates_directory(self, run_id):
-        return self._local_artifact_storage.intermediates_dir(run_id)
-
     def storage_directory(self):
         return self._local_artifact_storage.storage_dir
 
@@ -1345,6 +1409,11 @@ records = instance.get_event_records(
         from dagster.core.run_coordinator import SubmitRunContext
 
         run = self.get_run_by_id(run_id)
+        if run is None:
+            raise DagsterInvariantViolationError(
+                f"Could not load run {run_id} that was passed to submit_run"
+            )
+
         check.inst(
             run.external_pipeline_origin,
             ExternalPipelineOrigin,
@@ -1390,12 +1459,15 @@ records = instance.get_event_records(
             run_id (str): The id of the run the launch.
         """
         from dagster.core.launcher import LaunchRunContext
-
-        run = self.get_run_by_id(run_id)
-
         from dagster.core.events import EngineEventData, DagsterEvent, DagsterEventType
         from dagster.core.events.log import EventLogEntry
 
+        run = self.get_run_by_id(run_id)
+        if run is None:
+            raise DagsterInvariantViolationError(
+                f"Could not load run {run_id} that was passed to launch_run"
+            )
+
         launch_started_event = DagsterEvent(
             event_type_value=DagsterEventType.PIPELINE_STARTING.value,
             pipeline_name=run.pipeline_name,
@@ -1415,6 +1487,8 @@ records = instance.get_event_records(
         self.handle_new_event(event_record)
 
         run = self.get_run_by_id(run_id)
+        if run is None:
+            check.failed(f"Failed to reload run {run_id}")
 
         try:
             self._run_launcher.launch_run(LaunchRunContext(pipeline_run=run, workspace=workspace))
@@ -1430,6 +1504,67 @@ records = instance.get_event_records(
 
         return run
 
+    def resume_run(self, run_id: str, workspace: "IWorkspace", attempt_number: int):
+        """Resume a pipeline run.
+
+        This method should be called on runs which have already been launched, but whose run workers
+        have died.
+
+        Args:
+            run_id (str): The id of the run the launch.
+        """
+        from dagster.core.launcher import LaunchRunContext
+        from dagster.core.events import EngineEventData
+        from dagster.daemon.monitoring import RESUME_RUN_LOG_MESSAGE
+
+        run = self.get_run_by_id(run_id)
+        if run is None:
+            raise DagsterInvariantViolationError(
+                f"Could not load run {run_id} that was passed to resume_run"
+            )
+        if run.status not in IN_PROGRESS_RUN_STATUSES:
+            raise DagsterInvariantViolationError(
+                f"Run {run_id} is not in a state that can be resumed"
+            )
+
+        self.report_engine_event(
+            RESUME_RUN_LOG_MESSAGE,
+            run,
+        )
+
+        try:
+            self._run_launcher.launch_run(
+                LaunchRunContext(
+                    pipeline_run=run,
+                    workspace=workspace,
+                    resume_from_failure=True,
+                    resume_attempt_number=attempt_number,
+                )
+            )
+        except:
+            error = serializable_error_info_from_exc_info(sys.exc_info())
+            self.report_engine_event(
+                error.message,
+                run,
+                EngineEventData.engine_error(error),
+            )
+            self.report_run_failed(run)
+            raise
+
+        return run
+
+    def count_resume_run_attempts(self, run_id: str):
+        from dagster.daemon.monitoring import RESUME_RUN_LOG_MESSAGE
+        from dagster.core.events import DagsterEventType
+
+        events = self.all_logs(run_id, of_type=DagsterEventType.ENGINE_EVENT)
+        return len([event for event in events if event.message == RESUME_RUN_LOG_MESSAGE])
+
+    def run_will_resume(self, run_id: str):
+        if not self.run_monitoring_enabled:
+            return False
+        return self.count_resume_run_attempts(run_id) < self.run_monitoring_max_resume_run_attempts
+
     # Scheduler
 
     def reconcile_scheduler_state(self, external_repository):
@@ -1605,7 +1740,12 @@ records = instance.get_event_records(
     def get_required_daemon_types(self):
         from dagster.core.run_coordinator import QueuedRunCoordinator
         from dagster.core.scheduler import DagsterDaemonScheduler
-        from dagster.daemon.daemon import SchedulerDaemon, SensorDaemon, BackfillDaemon
+        from dagster.daemon.daemon import (
+            SchedulerDaemon,
+            SensorDaemon,
+            BackfillDaemon,
+            MonitoringDaemon,
+        )
         from dagster.daemon.run_coordinator.queued_run_coordinator_daemon import (
             QueuedRunCoordinatorDaemon,
         )
@@ -1618,6 +1758,8 @@ records = instance.get_event_records(
             daemons.append(SchedulerDaemon.daemon_type())
         if isinstance(self.run_coordinator, QueuedRunCoordinator):
             daemons.append(QueuedRunCoordinatorDaemon.daemon_type())
+        if self.run_monitoring_enabled:
+            daemons.append(MonitoringDaemon.daemon_type())
         return daemons
 
     # backfill
